from concurrent.futures import Executor
import array
import asyncio

import numpy as np
from rdp import rdp

from notool.utils import start_generator


class RdpDecimateSeries:
    def __init__(self, x_typecode: str = 'f', y_typecode: str = 'f', chunk_size: int = 1000,
                 epsilon=0.0, threshold=0.0, *, loop=None, pool=None):
        self.x = array.array(x_typecode)
        self.y = array.array(y_typecode)
        self.epsilon = epsilon
        self.threshold = threshold
        self.chunk_size = chunk_size
        self._force_processing = False
        self._freeze = False
        self._head_size = 0
        self._tail_size = 0
        self._in_process_size = 0
        self._loop: asyncio.AbstractEventLoop = loop
        self._pool: Executor = pool

    def copy(self, *, chunk_size=None, epsilon=None, threshold=None, freeze=False):
        object_copy = RdpDecimateSeries(self.x.typecode, self.y.typecode, chunk_size or self.chunk_size,
                                        epsilon or self.epsilon, threshold or self.threshold,
                                        loop=None if freeze else self._loop,
                                        pool=None if freeze else self._pool)
        object_copy.x = self.x
        object_copy.y = self.y
        if freeze:
            object_copy._freeze = True
        else:
            object_copy._tail_size = len(self.x)
        return object_copy

    def append(self, x, y):
        if not self._freeze:
            self.x.append(x)
            self.y.append(y)
            self._tail_size += 1
            self._check_for_processing()

    def extend(self, x, y):
        if not self._freeze:
            self.x.extend(x)
            self.y.extend(y)
            self._tail_size += len(x)
            self._check_for_processing()

    @property
    def freeze(self):
        return self._freeze and not self._in_process_size

    def processing_done(self):
        return self._loop.create_task(self._processing_done()) if self._loop else \
            not self._in_process_size

    # TODO реализовать другую логику ожидания завершения
    async def _processing_done(self):
        while self._in_process_size:
            await asyncio.sleep(0.1)

    def force_processing(self, freeze=True):
        if not self._freeze:
            self._force_processing = True
            self._freeze = freeze
            self._check_for_processing()

    def _check_for_processing(self):
        if not self._in_process_size and (self._tail_size >= self.chunk_size or self._force_processing):
            self._force_processing = False
            self._in_process_size = self._head_size + self._tail_size
            self._tail_size = 0
            args = (self.x, self.y, self._head_size, self._in_process_size,
                    self.epsilon, self.threshold)
            if self._loop:
                self._loop.create_task(self._processing_task(*args))
            else:
                self._processing_completion(*self._processing(*args))

    def _processing_completion(self, x, y):
        self._head_size = len(x)
        if self._tail_size:
            x.extend(self.x[-self._tail_size:])
            y.extend(self.y[-self._tail_size:])
        self.x = x
        self.y = y
        self._in_process_size = 0

    async def _processing_task(self, *args):
        x, y = await self._loop.run_in_executor(self._pool, self._processing, *args)
        self._processing_completion(x, y)
        if self._freeze:
            self._loop = None
            self._pool = None
        else:
            self._check_for_processing()

    @staticmethod
    def _processing(x, y, head_size, in_process_size, epsilon, threshold):
        @start_generator
        def threshold_generator(index):
            previous = y[index]
            while True:
                index = yield previous
                current = y[index]
                previous = current if abs(current - previous) > threshold else previous

        head_size = head_size - 2 if head_size >= 2 else head_size
        y_getitem = threshold_generator(head_size).send if threshold else y.__getitem__
        tail = ((x[index], y_getitem(index)) for index in range(head_size, in_process_size))

        tail = np.fromiter(tail, dtype=np.dtype((float, 2)))    # TODO dtype определять по коду в array.array
        tail = rdp(tail, epsilon=epsilon)
        x_tail, y_tail = np.split(tail.T, 2)
        x, y = x[:head_size], y[:head_size]
        x.extend(x_tail.flat)
        y.extend(y_tail.flat)
        return x, y

    def __len__(self):
        return len(self.x)


if __name__ == '__main__':

    def _tests():
        import matplotlib.pyplot as plt
        import math

        def factors_test():
            series = [(array.array('f', [0.00019972222798969597, 0.000755277753341943, 0.0018577778246253729, 0.004626666661351919, 0.008550277911126614, 0.011888055130839348, 0.012981944717466831, 0.01353750005364418, 0.014648610725998878, 0.015204166993498802, 0.0174263883382082, 0.019648611545562744, 0.020759722217917442, 0.021319445222616196, 0.023533055558800697, 0.024075554683804512, 0.0251736119389534, 0.025724999606609344, 0.0279905553907156, 0.028546111658215523, 0.03519111126661301, 0.03574666753411293, 0.04572499915957451, 0.046315278857946396, 0.05219194293022156, 0.05269972234964371, 0.1534550040960312, 0.15377166867256165, 5.564626693725586, 5.565243244171143, 8.676276206970215, 8.676836013793945, 8.681276321411133, 8.681822776794434, 8.692360877990723, 8.692920684814453, 8.701289176940918, 8.701845169067383, 8.710716247558594, 8.711236953735352, 8.719622611999512, 8.720182418823242, 8.720681190490723, 8.721814155578613, 8.722370147705078, 8.73181438446045, 8.73237419128418, 8.739635467529297, 8.740181922912598, 8.74846363067627, 8.74901008605957, 8.749588012695312, 8.75127124786377, 8.751832008361816, 8.760703086853027, 8.761259078979492, 8.769044876098633, 8.769596099853516, 8.780173301696777, 8.780707359313965, 8.790160179138184, 8.79072093963623, 8.798476219177246, 8.799036026000977, 8.800143241882324, 8.800716400146484, 8.810694694519043, 8.811297416687012, 8.822369575500488, 8.822908401489258, 8.832383155822754, 8.832938194274902, 8.834053993225098, 8.834626197814941, 8.842934608459473, 8.843484878540039, 8.853493690490723, 8.85408878326416, 8.8546142578125, 8.855164527893066, 8.865737915039062, 8.866236686706543, 8.876801490783691, 8.87736988067627, 8.886818885803223, 8.887392044067383, 8.888493537902832, 8.88904094696045, 8.89902400970459, 8.899587631225586, 8.911293029785156, 8.911823272705078, 8.920694351196289, 8.92125415802002, 8.92181396484375, 8.92235279083252, 8.93349838256836, 8.934062957763672, 8.94571590423584, 8.94627571105957, 8.954618453979492, 8.955169677734375, 8.955673217773438, 8.957408905029297, 8.957916259765625, 8.967348098754883, 8.967907905578613, 8.968463897705078, 8.969010353088379, 8.969592094421387, 8.970151901245117, 8.980152130126953, 8.980720520019531, 8.981280326843262, 8.981840133666992, 8.988511085510254, 8.989036560058594, 8.98962688446045, 8.991275787353516, 8.99178409576416, 9.001241683959961, 9.001784324645996, 9.002912521362305, 9.00407600402832, 9.004600524902344, 9.01404094696045, 9.014595985412598, 9.01512622833252, 9.016805648803711, 9.01736068725586, 9.026275634765625, 9.026826858520508, 9.02739143371582, 9.0296049118042, 9.030125617980957, 9.037925720214844, 9.038472175598145, 9.040130615234375, 9.040716171264648, 9.050737380981445, 9.051280975341797, 9.051810264587402, 9.053511619567871, 9.054032325744629, 9.066258430480957, 9.066797256469727, 9.075142860412598, 9.07567310333252, 9.076831817626953, 9.077369689941406, 9.090699195861816, 9.091233253479004, 9.103489875793457, 9.10404109954834, 9.116250038146973, 9.11681842803955, 9.130139350891113, 9.130681037902832, 9.144604682922363, 9.145125389099121, 9.156826972961426, 9.157374382019043, 9.157938003540039, 9.159565925598145, 9.16012191772461, 9.172334671020508, 9.172929763793945, 9.185112953186035, 9.185677528381348, 9.186814308166504, 9.187339782714844, 9.201275825500488, 9.201827049255371, 9.21347713470459, 9.214040756225586, 9.214588165283203, 9.215165138244629, 9.215672492980957, 9.21628475189209, 9.229561805725098, 9.230116844177246, 9.245702743530273, 9.246293067932129, 9.258493423461914, 9.25903606414795, 9.259588241577148, 9.260126113891602, 9.260685920715332, 9.261271476745605, 9.274596214294434, 9.275147438049316, 9.275690078735352, 9.276232719421387, 9.290724754333496, 9.291236877441406, 9.291784286499023, 9.292352676391602, 9.307374000549316, 9.307903289794922, 9.308450698852539, 9.309053421020508, 9.309600830078125, 9.310152053833008, 9.323490142822266, 9.324014663696289, 9.339591979980469, 9.340129852294922, 9.356276512145996, 9.35681438446045, 9.372386932373047, 9.37293815612793, 9.38904094696045, 9.389592170715332, 9.39123249053955, 9.391810417175293, 9.392334938049316, 9.405152320861816, 9.405712127685547, 9.406271934509277, 9.406826972961426, 9.407330513000488, 9.408472061157227, 9.40902328491211, 9.42125415802002, 9.421805381774902, 9.423494338989258, 9.424032211303711, 9.424570083618164, 9.43848991394043, 9.43904972076416, 9.43954849243164, 9.441263198852539, 9.441783905029297, 9.457921028137207, 9.458494186401367, 9.460134506225586, 9.460702896118164, 9.461254119873047, 9.474566459655762, 9.475160598754883, 9.475672721862793, 9.476232528686523, 9.476831436157227, 9.477912902832031, 9.478433609008789, 9.491787910461426, 9.492335319519043, 9.492886543273926, 9.494574546813965, 9.495147705078125, 9.511271476745605, 9.511792182922363, 9.512361526489258, 9.512899398803711, 9.513455390930176, 9.514005661010742, 9.531801223754883, 9.532326698303223, 9.533463478088379, 9.534574508666992, 9.535125732421875, 9.548433303833008, 9.548989295959473, 9.550108909606934, 9.551836013793945, 9.552356719970703, 9.569587707519531, 9.570096015930176, 9.570703506469727, 9.572347640991211, 9.572937965393066, 9.590108871459961, 9.590677261352539, 9.591211318969727, 9.592938423156738, 9.593481063842773, 9.609561920166016, 9.610116958618164, 9.630125999450684, 9.630672454833984, 9.63122844696045, 9.631827354431152, 9.63290786743164, 9.633437156677246, 9.650659561157227, 9.651227951049805, 9.651823043823242, 9.653997421264648, 9.654579162597656, 9.67567253112793, 9.676206588745117, 9.677912712097168, 9.678424835205078, 9.695143699645996, 9.695712089538574, 9.69624137878418, 9.696797370910645, 9.697908401489258, 9.69845962524414, 9.720151901245117, 9.720672607421875, 9.741823196411133, 9.742369651794434, 9.743480682373047, 9.744596481323242, 9.745134353637695, 9.765130043029785, 9.765689849853516, 9.767334938049316, 9.767877578735352, 9.768442153930664, 9.786775588989258, 9.787361145019531, 9.78792953491211, 9.78957462310791, 9.790139198303223, 9.790681838989258, 9.791250228881836, 9.814600944519043, 9.81513500213623, 9.81564712524414, 9.816246032714844, 9.835668563842773, 9.836211204528809, 9.83675765991211, 9.837352752685547, 9.838459014892578, 9.839015007019043, 9.86177921295166, 9.862357139587402, 9.86287784576416, 9.864006042480469, 9.864540100097656, 9.88960075378418, 9.890095710754395, 9.890689849853516, 9.891801834106445, 9.89236068725586, 9.892894744873047, 9.91510009765625, 9.915685653686523, 9.916214942932129, 9.917868614196777, 9.91843318939209, 9.920112609863281, 9.920646667480469, 9.921241760253906, 9.939592361450195, 9.940130233764648, 9.940690040588379, 9.942869186401367, 9.943485260009766, 9.971762657165527, 9.972312927246094, 9.973484992980957, 9.974023818969727, 9.99903678894043, 9.999557495117188, 10.00009536743164, 10.000699043273926, 10.001757621765137, 10.002334594726562, 10.024544715881348, 10.025116920471191, 10.025690078735352, 10.02733039855957, 10.027894973754883, 10.028424263000488, 10.029544830322266, 10.030113220214844, 10.030651092529297, 10.057339668273926, 10.057873725891113, 10.058442115783691, 10.060126304626465, 10.060694694519043, 10.089574813842773, 10.090112686157227, 10.091254234313965, 10.092326164245605, 10.09286880493164, 10.118437767028809, 10.119001388549805, 10.119574546813965, 10.120685577392578, 10.121241569519043, 10.150694847106934, 10.151224136352539, 10.151775360107422, 10.153424263000488, 10.154036521911621, 10.154540061950684, 10.155095100402832, 10.156254768371582, 10.15680980682373, 10.183476448059082, 10.184036254882812, 10.185160636901855, 10.187339782714844, 10.187899589538574, 10.18898868560791, 10.189535140991211, 10.19013500213623, 10.220685958862305, 10.221211433410645, 10.221814155578613, 10.223420143127441, 10.223993301391602, 10.252313613891602, 10.252890586853027, 10.253494262695312, 10.253979682922363, 10.2545747756958, 10.25621509552002, 10.256814002990723, 10.29065990447998, 10.29126262664795, 10.291740417480469, 10.294027328491211, 10.294591903686523, 10.323429107666016, 10.323975563049316, 10.324535369873047, 10.32731819152832, 10.327917098999023, 10.32957935333252, 10.330100059509277, 10.33065128326416, 10.331241607666016, 10.356215476989746, 10.356779098510742, 10.35732650756836, 10.360108375549316, 10.360642433166504, 10.361783981323242, 10.362899780273438, 10.363459587097168, 10.397356986999512, 10.397860527038574, 10.398463249206543, 10.401215553283691, 10.40178394317627, 10.403433799743652, 10.403992652893066, 10.438403129577637, 10.439005851745605, 10.4395751953125, 10.441736221313477, 10.44233512878418, 10.443398475646973, 10.44506549835205, 10.445672988891602, 10.479514122009277, 10.480125427246094, 10.480647087097168, 10.483450889587402, 10.483966827392578, 10.485130310058594, 10.48731803894043, 10.48787784576416, 10.488954544067383, 10.490646362304688, 10.4912109375, 10.520681381225586, 10.521198272705078, 10.521736145019531, 10.524565696716309, 10.5250825881958, 10.526224136352539, 10.527843475341797, 10.528424263000488, 10.569548606872559, 10.570130348205566, 10.57118034362793, 10.572860717773438, 10.573468208312988, 10.575112342834473, 10.57565975189209, 10.576197624206543, 10.618398666381836, 10.619006156921387, 10.620121955871582, 10.622335433959961, 10.62289047241211, 10.660099983215332, 10.660659790039062, 10.661775588989258, 10.663433074951172, 10.663962364196777, 10.665108680725098, 10.66564655303955, 10.666254043579102, 10.710108757019043, 10.71064281463623, 10.711202621459961, 10.713971138000488, 10.7145357131958, 10.715660095214844, 10.71618938446045, 10.716723442077637, 10.71788215637207, 10.767304420471191, 10.767855644226074, 10.768997192382812, 10.769505500793457, 10.770073890686035, 10.771193504333496, 10.772847175598145, 10.773398399353027, 10.774005889892578, 10.817309379577637, 10.81786060333252, 10.818415641784668, 10.8206205368042, 10.821188926696777, 10.823407173156738, 10.823949813842773, 10.82451343536377, 10.875624656677246, 10.876172065734863, 10.876740455627441, 10.879531860351562, 10.880061149597168, 10.881184577941895, 10.882278442382812, 10.882838249206543, 10.884491920471191, 10.885065078735352, 10.885607719421387, 10.940608024597168, 10.941176414489746, 10.941732406616211, 10.94233512878418, 10.942877769470215, 10.944509506225586, 10.945086479187012, 10.9467191696167, 10.94731330871582, 10.947830200195312, 11.002890586853027, 11.00342845916748, 11.003962516784668, 11.006233215332031, 11.006731986999512, 11.007278442382812, 11.007843017578125, 11.00841999053955, 11.009570121765137, 11.010090827941895, 11.063949584960938, 11.06451416015625, 11.065077781677246, 11.06727409362793, 11.067852020263672, 11.06896686553955, 11.070611953735352, 11.071736335754395, 11.072287559509277, 11.07287311553955, 11.132834434509277, 11.133394241333008, 11.13394546508789, 11.14006519317627, 11.140646934509277, 11.141753196716309, 11.142278671264648, 11.143406867980957, 11.143954277038574, 11.144553184509277, 11.146215438842773, 11.146732330322266, 11.147261619567871, 11.211189270019531, 11.211771011352539, 11.212300300598145, 11.214539527893066, 11.215065002441406, 11.216154098510742, 11.216740608215332, 11.218411445617676, 11.218962669372559, 11.28509521484375, 11.285637855529785, 11.286731719970703, 11.289544105529785, 11.290103912353516, 11.29064655303955, 11.29118537902832, 11.291709899902344, 11.292817115783691, 11.293380737304688, 11.295612335205078, 11.296223640441895, 11.296727180480957, 11.373984336853027, 11.374513626098633, 11.375052452087402, 11.377313613891602, 11.377864837646484, 11.378945350646973, 11.379491806030273, 11.381184577941895, 11.38174057006836, 11.382877349853516, 11.38338565826416, 11.384527206420898, 11.471757888793945, 11.472261428833008, 11.473442077636719, 11.473949432373047, 11.4744873046875, 11.475043296813965, 11.477317810058594, 11.477855682373047, 11.478415489196777, 11.558937072753906, 11.559514045715332, 11.560073852539062, 11.561710357666016, 11.562322616577148, 11.562872886657715, 11.563364028930664, 11.56392765045166, 11.565091133117676, 11.567852020263672, 11.568389892578125, 11.569497108459473, 11.570030212402344, 11.570642471313477, 11.572257041931152, 11.572808265686035, 11.573407173156738, 11.667864799499512, 11.668407440185547, 11.6689453125, 11.670646667480469, 11.671189308166504, 11.671744346618652, 11.67229175567627, 11.673419952392578, 11.675056457519531, 11.675633430480957, 11.6773042678833, 11.67782974243164, 11.678945541381836, 11.679488182067871, 11.780594825744629, 11.781167030334473, 11.781757354736328, 11.783407211303711, 11.783980369567871, 11.784531593322754, 11.785056114196777, 11.78616714477539, 11.787812232971191, 11.788415908813477, 11.788975715637207, 11.789475440979004, 11.790590286254883, 11.791749000549316, 11.792252540588379, 11.792847633361816, 11.7933988571167, 11.79451847076416, 11.795090675354004, 11.795598983764648, 11.796184539794922, 11.914501190185547, 11.91505241394043, 11.916146278381348, 11.918389320373535, 11.918936729431152, 11.920039176940918, 11.920598983764648, 11.92227840423584, 11.922808647155762, 11.92392349243164, 11.924513816833496, 11.925581932067871, 11.926141738891602, 11.92782974243164, 11.928372383117676, 11.92895793914795, 11.93450927734375, 11.935043334960938, 11.935572624206543, 12.019466400146484, 12.020021438598633, 12.020577430725098, 12.027791023254395, 12.028402328491211, 12.02894115447998, 12.032291412353516, 12.03283405303955, 12.033393859863281, 12.036176681518555, 12.03669261932373, 12.037274360656738, 12.0400390625, 12.040581703186035, 12.041714668273926, 12.0444917678833, 12.045061111450195, 12.046141624450684, 12.04836368560791, 12.048910140991211, 12.050043106079102, 12.050621032714844, 12.051674842834473, 12.052803993225098, 12.053971290588379, 12.054466247558594, 12.055624961853027, 12.056123733520508, 12.05668830871582, 12.17227840423584, 12.17283821105957, 12.173359870910645, 12.176145553588867, 12.176692962646484, 12.1773042678833, 12.180082321166992, 12.18056869506836, 12.181150436401367, 12.18447494506836, 12.185017585754395, 12.185590744018555, 12.188406944274902, 12.188923835754395, 12.189483642578125, 12.192265510559082, 12.192790985107422, 12.193949699401855, 12.196736335754395, 12.197248458862305, 12.197799682617188, 12.200594902038574, 12.201128959655762, 12.201692581176758, 12.204509735107422, 12.205060958862305, 12.20612907409668, 12.208380699157715, 12.208910942077637, 12.211670875549316, 12.212291717529297, 12.21279525756836, 12.213923454284668, 12.214491844177246, 12.215008735656738, 12.336701393127441, 12.337257385253906, 12.33779525756836, 12.345008850097656, 12.345602989196777, 12.346163749694824, 12.353367805480957, 12.353923797607422, 12.354475021362305, 12.357239723205566, 12.35779094696045, 12.358345985412598, 12.36111068725586, 12.36168384552002, 12.36278247833252, 12.363393783569336, 12.363954544067383, 12.365612030029297, 12.366158485412598, 12.366697311401367, 12.369453430175781, 12.370013236999512, 12.371158599853516, 12.373385429382324, 12.373885154724121, 12.375025749206543, 12.376128196716309, 12.376688003540039, 12.377838134765625, 12.378911018371582, 12.37946605682373, 12.380034446716309, 12.381141662597656, 12.381688117980957, 12.383380889892578, 12.383914947509766, 12.38447093963623, 12.453372955322266, 12.453906059265137, 12.454487800598145, 12.461731910705566, 12.462252616882324, 12.462812423706055, 12.470000267028809, 12.470599174499512, 12.471123695373535, 12.474457740783691, 12.47501277923584, 12.475564002990723, 12.476128578186035, 12.476675033569336, 12.478394508361816, 12.47891902923584, 12.479466438293457, 12.48277759552002, 12.483389854431152, 12.483919143676758, 12.486709594726562, 12.487265586853027, 12.488333702087402, 12.488893508911133, 12.489470481872559, 12.491154670715332, 12.491692543029785, 12.492803573608398, 12.49561595916748, 12.496167182922363, 12.496658325195312, 12.497274398803711, 12.497825622558594, 12.499444007873535, 12.5, 12.501141548156738, 12.501688003540039, 12.5022611618042, 12.50391960144043, 12.504487991333008, 12.505568504333496, 12.506141662597656, 12.507790565490723, 12.508341789245605, 12.5094575881958, 12.51001262664795, 12.510546684265137, 12.512239456176758, 12.512786865234375, 12.513936996459961, 12.516111373901367, 12.51666259765625, 12.517807960510254, 12.518346786499023, 12.518945693969727, 12.520607948303223, 12.52114200592041, 12.522265434265137, 12.52281665802002, 12.523398399353027, 12.525017738342285, 12.52556037902832, 12.52672290802002, 12.527243614196777, 12.528910636901855, 12.529475212097168, 12.531128883361816, 12.531688690185547, 12.533376693725586, 12.533919334411621, 12.535052299499512, 12.53555965423584, 12.53616714477539, 12.537230491638184, 12.5377779006958, 12.539487838745117, 12.540008544921875, 12.541666984558105, 12.542278289794922, 12.543354988098145, 12.543914794921875, 12.54445743560791, 12.545581817626953, 12.546154022216797, 12.548351287841797, 12.548910140991211, 12.550039291381836, 12.552278518676758, 12.55282974243164, 12.55388069152832, 12.55444049835205, 12.55668830871582, 12.557252883911133, 12.55783462524414, 12.561141967773438, 12.561701774597168, 12.56225299835205, 12.658880233764648, 12.659509658813477, 12.660043716430664, 12.667265892028809, 12.66779899597168, 12.668363571166992, 12.66891098022461, 12.669431686401367, 12.671679496765137, 12.672239303588867, 12.672781944274902, 12.673372268676758, 12.67388916015625, 12.675551414489746, 12.676119804382324, 12.67725658416748, 12.67782974243164, 12.6799955368042, 12.68060302734375, 12.681137084960938, 12.681684494018555, 12.682252883911133, 12.683915138244629, 12.684431076049805, 12.685546875, 12.686171531677246, 12.688355445861816, 12.688884735107422, 12.689449310302734, 12.689987182617188, 12.690555572509766, 12.692214012145996, 12.692773818969727, 12.694478988647461, 12.695030212402344, 12.696675300598145, 12.697265625, 12.69776439666748, 12.698376655578613, 12.698875427246094, 12.700542449951172, 12.701098442077637, 12.702248573303223, 12.702808380126953, 12.705008506774902, 12.705586433410645, 12.70610237121582, 12.706697463989258, 12.707230567932129, 12.708893775939941, 12.709492683410645, 12.711158752441406, 12.711714744567871, 12.71335506439209, 12.71390151977539, 12.714991569519043, 12.715577125549316, 12.717252731323242, 12.717799186706543, 12.719470977783203, 12.72002124786377, 12.721657752990723, 12.722261428833008, 12.723889350891113, 12.724435806274414, 12.725564002990723, 12.726141929626465, 12.727799415588379, 12.728333473205566, 12.730048179626465, 12.730603218078613, 12.732226371765137, 12.732786178588867, 12.733927726745605, 12.734478950500488, 12.736128807067871, 12.736658096313477, 12.738368034362793, 12.738910675048828, 12.740030288696289, 12.740572929382324, 12.742239952087402, 12.742803573608398, 12.744440078735352, 12.744996070861816, 12.746675491333008, 12.747218132019043, 12.748906135559082, 12.749465942382812, 12.75058650970459, 12.751145362854004, 12.752778053283691, 12.753341674804688, 12.753936767578125, 12.75446605682373, 12.755559921264648, 12.757218360900879, 12.75776481628418, 12.758932113647461, 12.759462356567383, 12.76111125946045, 12.761653900146484, 12.76223087310791, 12.762795448303223, 12.763910293579102, 12.765573501586914, 12.766132354736328, 12.767213821411133, 12.770012855529785, 12.77056884765625, 12.771123886108398, 12.77393627166748, 12.774435997009277, 12.775004386901855, 12.898367881774902, 12.898932456970215, 12.899466514587402, 12.902777671813965, 12.903346061706543, 12.903910636901855, 12.90665340423584, 12.907282829284668, 12.907812118530273, 12.910013198852539, 12.910547256469727, 12.911102294921875, 12.913923263549805, 12.914448738098145, 12.914996147155762, 12.915586471557617, 12.916146278381348, 12.918329238891602, 12.91891098022461, 12.919462203979492, 12.920064926147461, 12.920598983764648, 12.92332935333252, 12.923910140991211, 12.92443561553955, 12.926701545715332, 12.92723560333252, 12.92778205871582, 12.928333282470703, 12.928884506225586, 12.931106567382812, 12.931671142578125, 12.932795524597168, 12.933359146118164, 12.935009002685547, 12.935546875, 12.936697006225586, 12.937226295471191, 12.939478874206543, 12.940021514892578, 12.941145896911621, 12.9417142868042, 12.94332504272461, 12.943888664245605, 12.94443130493164, 12.944999694824219, 12.945598602294922, 12.947782516479492, 12.948328971862793, 12.949475288391113, 12.950043678283691, 12.951680183410645, 12.952230453491211, 12.952799797058105, 12.953390121459961, 12.953919410705566, 12.956097602844238, 12.956679344177246, 12.957781791687012, 12.958372116088867, 12.959996223449707, 12.960542678833008, 12.961163520812988, 12.961662292480469, 12.962812423706055, 12.96447467803955, 12.965030670166016, 12.966133117675781, 12.966706275939941, 12.968341827392578, 12.968897819519043, 12.969470977783203, 12.970008850097656, 12.971137046813965, 12.972773551940918, 12.973341941833496, 12.974466323852539, 12.975021362304688, 12.976705551147461, 12.977256774902344, 12.977760314941406, 12.978311538696289, 12.979470252990723, 12.981154441833496, 12.981653213500977, 12.983359336853027, 12.983927726745605, 12.985013008117676, 12.985589981079102, 12.986141204833984, 12.986701011657715, 12.987812042236328, 12.989422798156738, 12.990004539489746, 12.991714477539062, 12.992243766784668, 12.993364334106445, 12.993919372558594, 12.995559692382812, 12.996110916137695, 12.997760772705078, 12.99834156036377, 13.000030517578125, 13.000534057617188, 13.001714706420898, 13.002257347106934, 13.003867149353027, 13.00443172454834, 13.006089210510254, 13.006714820861816, 13.008355140686035, 13.008910179138184, 13.010560035705566, 13.011124610900879, 13.012764930725098, 13.01332950592041, 13.014435768127441, 13.015000343322754, 13.01557731628418, 13.01610279083252, 13.017239570617676, 13.018311500549316, 13.018918991088867, 13.021089553833008, 13.021714210510254, 13.022786140441895, 13.023906707763672, 13.024435997009277, 13.025594711303711, 13.026102066040039, 13.029439926147461, 13.029991149902344, 13.03109359741211, 13.031697273254395, 13.034431457519531, 13.034996032714844, 13.035581588745117, 13.038923263549805, 13.039475440979004, 13.040034294128418, 13.042213439941406, 13.042764663696289, 13.043329238891602, 13.142195701599121, 13.142765045166016, 13.143316268920898, 13.146120071411133, 13.146736145019531, 13.147256851196289, 13.14998722076416, 13.15054702758789, 13.151688575744629, 13.153328895568848, 13.153910636901855, 13.154457092285156, 13.154987335205078, 13.155550956726074, 13.15835952758789, 13.158863067626953, 13.160035133361816, 13.161649703979492, 13.162243843078613, 13.162764549255371, 13.163363456726074, 13.164453506469727, 13.166132926940918, 13.166687965393066, 13.167218208312988, 13.167756080627441, 13.168368339538574, 13.170008659362793, 13.170551300048828, 13.171097755432129, 13.171723365783691, 13.172795295715332, 13.17443561553955, 13.174991607666016, 13.175559997558594, 13.176107406616211, 13.17723560333252, 13.179466247558594, 13.179973602294922, 13.181111335754395, 13.18224811553955, 13.182790756225586, 13.183372497558594, 13.183867454528809, 13.184483528137207, 13.185599327087402, 13.186089515686035, 13.186671257019043, 13.187777519226074, 13.188337326049805, 13.188897132873535, 13.1899995803833, 13.190589904785156, 13.193897247314453, 13.194422721862793, 13.194995880126953, 13.197782516479492, 13.198320388793945, 13.198880195617676, 13.202221870422363, 13.20276927947998, 13.203311920166016, 13.206132888793945, 13.206687927246094, 13.207221984863281, 13.457209587097168, 13.457808494567871, 13.458351135253906, 13.478328704833984, 13.478897094726562, 13.479453086853027, 13.482287406921387, 13.48279094696045, 13.483880043029785, 13.485581398010254, 13.486084938049316, 13.486692428588867, 13.487222671508789, 13.487751960754395, 13.49052906036377, 13.49111557006836, 13.49223518371582, 13.493866920471191, 13.494431495666504, 13.494996070861816, 13.495555877685547, 13.496115684509277, 13.498888969421387, 13.499452590942383, 13.500568389892578, 13.502213478088379, 13.50277328491211, 13.503324508666992, 13.50395393371582, 13.504414558410645, 13.5061674118042, 13.506692886352539, 13.507247924804688, 13.507756233215332, 13.508880615234375, 13.510581970214844, 13.51111125946045, 13.511674880981445, 13.512239456176758, 13.51332950592041, 13.515555381774902, 13.516119956970215, 13.517226219177246, 13.518324851989746, 13.518884658813477, 13.519444465637207, 13.52004337310791, 13.520564079284668, 13.521679878234863, 13.522799491882324, 13.523859024047852, 13.524409294128418, 13.525581359863281, 13.526115417480469, 13.526688575744629, 13.527253150939941, 13.527782440185547, 13.528350830078125, 13.528910636901855, 13.5316801071167, 13.532248497009277, 13.53278636932373, 13.53388500213623, 13.534444808959961, 13.535542488098145, 13.536124229431152, 13.537755966186523, 13.53830337524414, 13.539401054382324, 13.540568351745605, 13.54112434387207, 13.542200088500977, 13.542769432067871, 13.54332447052002, 13.543880462646484, 13.544501304626465, 13.546089172363281, 13.546653747558594, 13.547781944274902, 13.548341751098633, 13.852768898010254, 13.85329818725586, 13.853876113891602, 13.88219165802002, 13.88275146484375, 13.883281707763672, 13.887166023254395, 13.887786865234375, 13.888311386108398, 13.892195701599121, 13.892743110656738, 13.893342018127441, 13.897226333618164, 13.897782325744629, 13.898324966430664, 13.902231216430664, 13.902791023254395, 13.903862953186035, 13.906614303588867, 13.90717887878418, 13.908342361450195, 13.908893585205078, 13.911102294921875, 13.911636352539062, 13.91221809387207, 13.912769317626953, 13.913849830627441, 13.917174339294434, 13.917738914489746, 13.91883659362793, 13.922218322753906, 13.92279052734375, 13.92386245727539, 13.924423217773438, 13.924983024597168, 13.925525665283203, 13.92720890045166, 13.927764892578125, 13.928884506225586, 13.929405212402344, 13.932747840881348, 13.933316230773926, 13.934440612792969, 13.934978485107422, 13.936640739440918, 13.937200546264648, 13.937765121459961, 13.938328742980957, 13.93941879272461, 13.939982414245605, 13.941679954528809, 13.942214012145996, 13.942764282226562, 13.943876266479492, 13.944952011108398, 13.945542335510254, 13.946102142333984, 13.946666717529297, 13.947165489196777, 13.94830322265625, 13.948837280273438, 13.951072692871094, 13.95167064666748, 13.95276927947998, 13.953351020812988, 13.953880310058594, 13.955516815185547, 13.95610237121582, 13.956640243530273, 13.957786560058594, 13.958307266235352, 13.961653709411621, 13.96220874786377, 13.96273422241211, 13.963862419128418, 13.96444034576416, 13.967734336853027, 13.96829891204834, 13.969430923461914, 13.97164535522461, 13.972230911254883, 13.972734451293945, 13.977782249450684, 13.978294372558594, 13.978862762451172, 13.987231254577637, 13.987773895263672, 13.988316535949707, 13.992755889892578, 13.99332046508789, 13.993880271911621, 14.144461631774902]),
                       array.array('f', [0.0, 1.399999976158142, 0.4399999976158142, 0.27000001072883606, 0.20999999344348907, 0.18000000715255737, 0.18000000715255737, 0.17000000178813934, 0.17000000178813934, 0.1599999964237213, 0.1599999964237213, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.0, 0.0, -1.159999966621399, -1.159999966621399, 1.399999976158142, 1.399999976158142, 1.3899999856948853, 1.3899999856948853, 1.3799999952316284, 1.3799999952316284, 1.3700000047683716, 1.3700000047683716, 1.3600000143051147, 1.3600000143051147, 1.350000023841858, 1.350000023841858, 1.340000033378601, 1.350000023841858, 1.350000023841858, 1.340000033378601, 1.340000033378601, 1.3300000429153442, 1.3300000429153442, 1.3200000524520874, 1.3200000524520874, 1.309999942779541, 1.3200000524520874, 1.3200000524520874, 1.309999942779541, 1.309999942779541, 1.2999999523162842, 1.2999999523162842, 1.2899999618530273, 1.2899999618530273, 1.2799999713897705, 1.2799999713897705, 1.2699999809265137, 1.2699999809265137, 1.2599999904632568, 1.2699999809265137, 1.2599999904632568, 1.2599999904632568, 1.25, 1.25, 1.2400000095367432, 1.2400000095367432, 1.2300000190734863, 1.2400000095367432, 1.2300000190734863, 1.2300000190734863, 1.2200000286102295, 1.2200000286102295, 1.2100000381469727, 1.2200000286102295, 1.2100000381469727, 1.2100000381469727, 1.2000000476837158, 1.2000000476837158, 1.190000057220459, 1.190000057220459, 1.1799999475479126, 1.190000057220459, 1.1799999475479126, 1.1799999475479126, 1.1699999570846558, 1.1699999570846558, 1.159999966621399, 1.159999966621399, 1.149999976158142, 1.159999966621399, 1.149999976158142, 1.149999976158142, 1.1399999856948853, 1.1399999856948853, 1.1299999952316284, 1.1299999952316284, 1.1200000047683716, 1.1299999952316284, 1.1299999952316284, 1.1200000047683716, 1.1200000047683716, 1.1100000143051147, 1.1200000047683716, 1.1100000143051147, 1.1200000047683716, 1.1100000143051147, 1.1100000143051147, 1.100000023841858, 1.1100000143051147, 1.100000023841858, 1.100000023841858, 1.090000033378601, 1.100000023841858, 1.100000023841858, 1.090000033378601, 1.090000033378601, 1.0800000429153442, 1.090000033378601, 1.090000033378601, 1.0800000429153442, 1.0800000429153442, 1.0700000524520874, 1.0800000429153442, 1.0800000429153442, 1.0700000524520874, 1.0700000524520874, 1.059999942779541, 1.0700000524520874, 1.0700000524520874, 1.059999942779541, 1.059999942779541, 1.0499999523162842, 1.059999942779541, 1.0499999523162842, 1.0499999523162842, 1.0399999618530273, 1.0499999523162842, 1.0499999523162842, 1.0399999618530273, 1.0399999618530273, 1.0299999713897705, 1.0299999713897705, 1.0199999809265137, 1.0299999713897705, 1.0199999809265137, 1.0199999809265137, 1.0099999904632568, 1.0099999904632568, 1.0, 1.0, 0.9900000095367432, 0.9900000095367432, 0.9800000190734863, 0.9800000190734863, 0.9700000286102295, 0.9700000286102295, 0.9599999785423279, 0.9700000286102295, 0.9700000286102295, 0.9599999785423279, 0.9599999785423279, 0.949999988079071, 0.949999988079071, 0.9399999976158142, 0.949999988079071, 0.9399999976158142, 0.9399999976158142, 0.9300000071525574, 0.9300000071525574, 0.9200000166893005, 0.9300000071525574, 0.9200000166893005, 0.9300000071525574, 0.9200000166893005, 0.9200000166893005, 0.9100000262260437, 0.9100000262260437, 0.8999999761581421, 0.8999999761581421, 0.8899999856948853, 0.8999999761581421, 0.8899999856948853, 0.8999999761581421, 0.8899999856948853, 0.8899999856948853, 0.8799999952316284, 0.8899999856948853, 0.8799999952316284, 0.8799999952316284, 0.8700000047683716, 0.8799999952316284, 0.8700000047683716, 0.8700000047683716, 0.8600000143051147, 0.8700000047683716, 0.8600000143051147, 0.8700000047683716, 0.8600000143051147, 0.8600000143051147, 0.8500000238418579, 0.8500000238418579, 0.8399999737739563, 0.8399999737739563, 0.8299999833106995, 0.8299999833106995, 0.8199999928474426, 0.8199999928474426, 0.8100000023841858, 0.8100000023841858, 0.8199999928474426, 0.8100000023841858, 0.8100000023841858, 0.800000011920929, 0.8100000023841858, 0.800000011920929, 0.8100000023841858, 0.8100000023841858, 0.800000011920929, 0.800000011920929, 0.7900000214576721, 0.7900000214576721, 0.800000011920929, 0.7900000214576721, 0.7900000214576721, 0.7799999713897705, 0.7900000214576721, 0.7900000214576721, 0.7799999713897705, 0.7799999713897705, 0.7699999809265137, 0.7699999809265137, 0.7799999713897705, 0.7699999809265137, 0.7699999809265137, 0.7599999904632568, 0.7699999809265137, 0.7599999904632568, 0.7699999809265137, 0.7699999809265137, 0.7599999904632568, 0.7599999904632568, 0.75, 0.7599999904632568, 0.7599999904632568, 0.75, 0.75, 0.7400000095367432, 0.75, 0.7400000095367432, 0.75, 0.7400000095367432, 0.7400000095367432, 0.7300000190734863, 0.7300000190734863, 0.7400000095367432, 0.7300000190734863, 0.7300000190734863, 0.7200000286102295, 0.7300000190734863, 0.7300000190734863, 0.7200000286102295, 0.7200000286102295, 0.7099999785423279, 0.7200000286102295, 0.7200000286102295, 0.7099999785423279, 0.7099999785423279, 0.699999988079071, 0.7099999785423279, 0.7099999785423279, 0.699999988079071, 0.699999988079071, 0.6899999976158142, 0.6899999976158142, 0.6800000071525574, 0.6899999976158142, 0.6800000071525574, 0.6899999976158142, 0.6800000071525574, 0.6800000071525574, 0.6700000166893005, 0.6800000071525574, 0.6800000071525574, 0.6700000166893005, 0.6700000166893005, 0.6600000262260437, 0.6700000166893005, 0.6600000262260437, 0.6600000262260437, 0.6499999761581421, 0.6600000262260437, 0.6499999761581421, 0.6600000262260437, 0.6499999761581421, 0.6499999761581421, 0.6399999856948853, 0.6399999856948853, 0.6299999952316284, 0.6399999856948853, 0.6399999856948853, 0.6299999952316284, 0.6299999952316284, 0.6200000047683716, 0.6200000047683716, 0.6299999952316284, 0.6200000047683716, 0.6200000047683716, 0.6100000143051147, 0.6200000047683716, 0.6200000047683716, 0.6100000143051147, 0.6200000047683716, 0.6100000143051147, 0.6100000143051147, 0.6000000238418579, 0.6100000143051147, 0.6000000238418579, 0.6000000238418579, 0.5899999737739563, 0.6000000238418579, 0.5899999737739563, 0.6000000238418579, 0.5899999737739563, 0.5899999737739563, 0.5799999833106995, 0.5899999737739563, 0.5899999737739563, 0.5799999833106995, 0.5799999833106995, 0.5699999928474426, 0.5799999833106995, 0.5699999928474426, 0.5799999833106995, 0.5699999928474426, 0.5699999928474426, 0.5600000023841858, 0.5699999928474426, 0.5699999928474426, 0.5600000023841858, 0.5600000023841858, 0.5699999928474426, 0.5600000023841858, 0.5600000023841858, 0.550000011920929, 0.5600000023841858, 0.5600000023841858, 0.550000011920929, 0.550000011920929, 0.5400000214576721, 0.550000011920929, 0.5400000214576721, 0.5400000214576721, 0.5299999713897705, 0.5400000214576721, 0.5299999713897705, 0.5400000214576721, 0.5299999713897705, 0.5299999713897705, 0.5199999809265137, 0.5299999713897705, 0.5299999713897705, 0.5199999809265137, 0.5299999713897705, 0.5199999809265137, 0.5299999713897705, 0.5199999809265137, 0.5199999809265137, 0.5099999904632568, 0.5199999809265137, 0.5199999809265137, 0.5099999904632568, 0.5099999904632568, 0.5, 0.5099999904632568, 0.5099999904632568, 0.5, 0.5, 0.49000000953674316, 0.5, 0.5, 0.49000000953674316, 0.49000000953674316, 0.47999998927116394, 0.49000000953674316, 0.49000000953674316, 0.47999998927116394, 0.49000000953674316, 0.47999998927116394, 0.49000000953674316, 0.47999998927116394, 0.47999998927116394, 0.4699999988079071, 0.47999998927116394, 0.47999998927116394, 0.4699999988079071, 0.4699999988079071, 0.47999998927116394, 0.4699999988079071, 0.4699999988079071, 0.46000000834465027, 0.4699999988079071, 0.4699999988079071, 0.46000000834465027, 0.46000000834465027, 0.44999998807907104, 0.46000000834465027, 0.44999998807907104, 0.46000000834465027, 0.46000000834465027, 0.44999998807907104, 0.44999998807907104, 0.4399999976158142, 0.44999998807907104, 0.44999998807907104, 0.4399999976158142, 0.4399999976158142, 0.4300000071525574, 0.4399999976158142, 0.4399999976158142, 0.4300000071525574, 0.4399999976158142, 0.4300000071525574, 0.4399999976158142, 0.4300000071525574, 0.4300000071525574, 0.41999998688697815, 0.4300000071525574, 0.4300000071525574, 0.41999998688697815, 0.4300000071525574, 0.4300000071525574, 0.41999998688697815, 0.41999998688697815, 0.4099999964237213, 0.41999998688697815, 0.41999998688697815, 0.4099999964237213, 0.41999998688697815, 0.4099999964237213, 0.4099999964237213, 0.4000000059604645, 0.4099999964237213, 0.4099999964237213, 0.4000000059604645, 0.4099999964237213, 0.4099999964237213, 0.4000000059604645, 0.4000000059604645, 0.38999998569488525, 0.4000000059604645, 0.4000000059604645, 0.38999998569488525, 0.4000000059604645, 0.4000000059604645, 0.38999998569488525, 0.4000000059604645, 0.4000000059604645, 0.38999998569488525, 0.38999998569488525, 0.3799999952316284, 0.38999998569488525, 0.38999998569488525, 0.3799999952316284, 0.38999998569488525, 0.38999998569488525, 0.3799999952316284, 0.3799999952316284, 0.3700000047683716, 0.3799999952316284, 0.3799999952316284, 0.3700000047683716, 0.3700000047683716, 0.3799999952316284, 0.3700000047683716, 0.3700000047683716, 0.36000001430511475, 0.3700000047683716, 0.3700000047683716, 0.36000001430511475, 0.36000001430511475, 0.3499999940395355, 0.36000001430511475, 0.36000001430511475, 0.3499999940395355, 0.3499999940395355, 0.36000001430511475, 0.3499999940395355, 0.3499999940395355, 0.3400000035762787, 0.3499999940395355, 0.3499999940395355, 0.3400000035762787, 0.3499999940395355, 0.3400000035762787, 0.3499999940395355, 0.3400000035762787, 0.3400000035762787, 0.33000001311302185, 0.3400000035762787, 0.33000001311302185, 0.3400000035762787, 0.33000001311302185, 0.33000001311302185, 0.3400000035762787, 0.33000001311302185, 0.33000001311302185, 0.3199999928474426, 0.33000001311302185, 0.33000001311302185, 0.3199999928474426, 0.3199999928474426, 0.33000001311302185, 0.3199999928474426, 0.3199999928474426, 0.3100000023841858, 0.3199999928474426, 0.3199999928474426, 0.3100000023841858, 0.3199999928474426, 0.3199999928474426, 0.3100000023841858, 0.3100000023841858, 0.3199999928474426, 0.3100000023841858, 0.3100000023841858, 0.30000001192092896, 0.3100000023841858, 0.30000001192092896, 0.3100000023841858, 0.3100000023841858, 0.30000001192092896, 0.30000001192092896, 0.3100000023841858, 0.30000001192092896, 0.30000001192092896, 0.28999999165534973, 0.30000001192092896, 0.30000001192092896, 0.28999999165534973, 0.30000001192092896, 0.28999999165534973, 0.30000001192092896, 0.30000001192092896, 0.28999999165534973, 0.28999999165534973, 0.2800000011920929, 0.28999999165534973, 0.28999999165534973, 0.2800000011920929, 0.28999999165534973, 0.28999999165534973, 0.2800000011920929, 0.28999999165534973, 0.2800000011920929, 0.2800000011920929, 0.27000001072883606, 0.2800000011920929, 0.2800000011920929, 0.27000001072883606, 0.27000001072883606, 0.2800000011920929, 0.27000001072883606, 0.2800000011920929, 0.27000001072883606, 0.27000001072883606, 0.2800000011920929, 0.27000001072883606, 0.27000001072883606, 0.25999999046325684, 0.27000001072883606, 0.27000001072883606, 0.25999999046325684, 0.25999999046325684, 0.27000001072883606, 0.27000001072883606, 0.25999999046325684, 0.25999999046325684, 0.25, 0.25999999046325684, 0.25999999046325684, 0.25, 0.25999999046325684, 0.25, 0.25999999046325684, 0.25999999046325684, 0.25, 0.25, 0.25999999046325684, 0.25, 0.25, 0.23999999463558197, 0.25, 0.25, 0.23999999463558197, 0.23999999463558197, 0.25, 0.25, 0.23999999463558197, 0.23999999463558197, 0.25, 0.23999999463558197, 0.23999999463558197, 0.23000000417232513, 0.23999999463558197, 0.23000000417232513, 0.23999999463558197, 0.23000000417232513, 0.23000000417232513, 0.23999999463558197, 0.23000000417232513, 0.23000000417232513, 0.2199999988079071, 0.23000000417232513, 0.23000000417232513, 0.2199999988079071, 0.23000000417232513, 0.2199999988079071, 0.23000000417232513, 0.2199999988079071, 0.2199999988079071, 0.23000000417232513, 0.2199999988079071, 0.23000000417232513, 0.2199999988079071, 0.2199999988079071, 0.23000000417232513, 0.2199999988079071, 0.2199999988079071, 0.20999999344348907, 0.2199999988079071, 0.2199999988079071, 0.20999999344348907, 0.2199999988079071, 0.20999999344348907, 0.2199999988079071, 0.2199999988079071, 0.20999999344348907, 0.20999999344348907, 0.2199999988079071, 0.2199999988079071, 0.20999999344348907, 0.20999999344348907, 0.20000000298023224, 0.20999999344348907, 0.20999999344348907, 0.20000000298023224, 0.20999999344348907, 0.20000000298023224, 0.20999999344348907, 0.20999999344348907, 0.20000000298023224, 0.20999999344348907, 0.20000000298023224, 0.20999999344348907, 0.20999999344348907, 0.20000000298023224, 0.20999999344348907, 0.20000000298023224, 0.20999999344348907, 0.20000000298023224, 0.20999999344348907, 0.20000000298023224, 0.20000000298023224, 0.1899999976158142, 0.20000000298023224, 0.20000000298023224, 0.1899999976158142, 0.1899999976158142, 0.20000000298023224, 0.20000000298023224, 0.1899999976158142, 0.1899999976158142, 0.20000000298023224, 0.20000000298023224, 0.1899999976158142, 0.1899999976158142, 0.20000000298023224, 0.1899999976158142, 0.1899999976158142, 0.20000000298023224, 0.1899999976158142, 0.1899999976158142, 0.18000000715255737, 0.1899999976158142, 0.1899999976158142, 0.18000000715255737, 0.1899999976158142, 0.1899999976158142, 0.18000000715255737, 0.1899999976158142, 0.1899999976158142, 0.18000000715255737, 0.1899999976158142, 0.1899999976158142, 0.18000000715255737, 0.1899999976158142, 0.1899999976158142, 0.18000000715255737, 0.1899999976158142, 0.1899999976158142, 0.18000000715255737, 0.18000000715255737, 0.1899999976158142, 0.18000000715255737, 0.1899999976158142, 0.18000000715255737, 0.1899999976158142, 0.18000000715255737, 0.1899999976158142, 0.18000000715255737, 0.18000000715255737, 0.17000000178813934, 0.18000000715255737, 0.18000000715255737, 0.17000000178813934, 0.18000000715255737, 0.18000000715255737, 0.17000000178813934, 0.18000000715255737, 0.18000000715255737, 0.17000000178813934, 0.18000000715255737, 0.18000000715255737, 0.17000000178813934, 0.18000000715255737, 0.18000000715255737, 0.17000000178813934, 0.18000000715255737, 0.18000000715255737, 0.17000000178813934, 0.18000000715255737, 0.18000000715255737, 0.17000000178813934, 0.18000000715255737, 0.18000000715255737, 0.17000000178813934, 0.18000000715255737, 0.18000000715255737, 0.17000000178813934, 0.17000000178813934, 0.18000000715255737, 0.17000000178813934, 0.17000000178813934, 0.18000000715255737, 0.17000000178813934, 0.17000000178813934, 0.1599999964237213, 0.17000000178813934, 0.17000000178813934, 0.1599999964237213, 0.17000000178813934, 0.17000000178813934, 0.1599999964237213, 0.17000000178813934, 0.17000000178813934, 0.1599999964237213, 0.17000000178813934, 0.17000000178813934, 0.1599999964237213, 0.17000000178813934, 0.1599999964237213, 0.17000000178813934, 0.17000000178813934, 0.1599999964237213, 0.17000000178813934, 0.17000000178813934, 0.1599999964237213, 0.17000000178813934, 0.17000000178813934, 0.1599999964237213, 0.17000000178813934, 0.17000000178813934, 0.1599999964237213, 0.17000000178813934, 0.1599999964237213, 0.17000000178813934, 0.1599999964237213, 0.17000000178813934, 0.1599999964237213, 0.1599999964237213, 0.17000000178813934, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.15000000596046448, 0.1599999964237213, 0.15000000596046448, 0.15000000596046448, 0.1599999964237213, 0.1599999964237213, 0.15000000596046448, 0.15000000596046448, 0.1599999964237213, 0.15000000596046448, 0.15000000596046448, 0.1599999964237213, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.15000000596046448, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.14000000059604645, 0.12999999523162842, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.11999999731779099, 0.12999999523162842, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.10999999940395355, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.11999999731779099, 0.10999999940395355, 0.10999999940395355, 0.11999999731779099, 0.11999999731779099, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612, 0.10999999940395355, 0.10000000149011612, 0.10000000149011612])
                       ),
                      (array.array('f', [0.00019972222798969597, 0.0013022222556173801, 0.05219194293022156, 0.053797777742147446, 0.05490027740597725, 0.05878055468201637, 0.06212250143289566, 0.06327249854803085, 0.0666188895702362, 0.06713555753231049, 0.06880638748407364, 0.06989583373069763, 0.07493499666452408, 0.07549916952848434, 0.07823361456394196, 0.08045139163732529, 0.08825527876615524, 0.09103749692440033, 0.10545583069324493, 0.10878055542707443, 0.12384111434221268, 0.12716583907604218, 0.13937944173812866, 0.1399044394493103, 0.14932721853256226, 0.1534550040960312, 0.15434889495372772, 0.16212250292301178, 0.16877610981464386, 0.17047305405139923, 0.17099833488464355, 0.17268222570419312, 0.1754775047302246, 0.19934888184070587, 0.19989167153835297, 0.21102860569953918, 0.21159721910953522, 0.4254472255706787, 0.4260374903678894, 0.4654383361339569, 0.4660155475139618, 0.5043619275093079, 0.5048916935920715, 0.5832291841506958, 0.5837458372116089, 0.6526650190353394, 0.6532466411590576, 0.678754448890686, 0.6793055534362793, 0.713789165019989, 0.7143141627311707, 0.777126669883728, 0.7776474952697754, 0.8443055748939514, 0.8449088931083679, 0.8665669560432434, 0.8670833110809326, 0.8965322375297546, 0.8971094489097595, 0.9498480558395386, 0.9504383206367493, 1.0021225214004517, 1.002643346786499, 1.0221225023269653, 1.0226258039474487, 1.0498656034469604, 1.0504472255706787, 1.0965710878372192, 1.0970877408981323, 1.1581858396530151, 1.15876305103302, 1.1798917055130005, 1.1804035902023315, 1.2026258707046509, 1.2032119035720825, 1.250950574874878, 1.2515192031860352, 1.304861068725586, 1.305459976196289, 1.3271050453186035, 1.3276258707046509, 1.352091908454895, 1.352656364440918, 1.3926433324813843, 1.3931727409362793, 1.4492883682250977, 1.4498263597488403, 1.4726519584655762, 1.4732249975204468, 1.4943230152130127, 1.4948525428771973, 1.5348827838897705, 1.5353994369506836, 1.568732738494873, 1.5693230628967285, 1.5865236520767212, 1.5870963335037231, 1.6098308563232422, 1.6104124784469604, 1.6543231010437012, 1.6548349857330322, 1.7159852981567383, 1.7165061235427856, 1.7359460592269897, 1.736532211303711, 1.7592753171920776, 1.759826421737671, 1.7987500429153442, 1.7992666959762573, 1.8481857776641846, 1.8487282991409302, 1.8670616149902344, 1.8676258325576782, 1.8887499570846558, 1.8892794847488403, 1.9293099641799927, 1.9298394918441772, 1.9815192222595215, 1.9821008443832397, 2.0015363693237305, 2.002039909362793, 2.0237414836883545, 2.02427077293396, 2.0653862953186035, 2.065976619720459, 2.1082074642181396, 2.1087586879730225, 2.127039909362793, 2.127621650695801, 2.1493141651153564, 2.14984393119812, 2.1842708587646484, 2.1848177909851074, 2.2431554794311523, 2.243736982345581, 2.261488914489746, 2.2620701789855957, 2.283689260482788, 2.2842838764190674, 2.3220746517181396, 2.322604179382324, 2.365347146987915, 2.3659420013427734, 2.3859071731567383, 2.386492967605591, 2.405381917953491, 2.405907154083252, 2.447061777114868, 2.447582483291626, 2.4948394298553467, 2.495377779006958, 2.5131771564483643, 2.513698101043701, 2.5320355892181396, 2.532573938369751, 2.571484327316284, 2.572066068649292, 2.6037023067474365, 2.60426664352417, 2.620946168899536, 2.621466875076294, 2.642035484313965, 2.642578125, 2.6837282180786133, 2.68424916267395, 2.7314584255218506, 2.73205304145813, 2.7475781440734863, 2.7481555938720703, 2.76924467086792, 2.7698349952697754, 2.804804801940918, 2.805347204208374, 2.849236011505127, 2.8498222827911377, 2.867013931274414, 2.867595672607422, 2.8853647708892822, 2.8859288692474365, 2.9197874069213867, 2.920377731323242, 2.9670355319976807, 2.9676172733306885, 2.984778642654419, 2.985347270965576, 3.0036981105804443, 3.00422739982605, 3.0392751693725586, 3.0398309230804443, 3.077604055404663, 3.0781424045562744, 3.0942447185516357, 3.0947656631469727, 3.113715171813965, 3.114232063293457, 3.148151159286499, 3.1486804485321045, 3.1942319869995117, 3.1948349475860596, 3.2114453315734863, 3.212026834487915, 3.2286980152130127, 3.229236125946045, 3.2637112140655518, 3.2642273902893066, 3.3031208515167236, 3.3037021160125732, 3.3175477981567383, 3.3181424140930176, 3.335902690887451, 3.3364410400390625, 3.367595672607422, 3.368133544921875, 3.40871524810791, 3.4092319011688232, 3.426475763320923, 3.4270529747009277, 3.4419877529144287, 3.4425954818725586, 3.4748003482818604, 3.475330114364624, 3.504253625869751, 3.504822254180908, 3.51921010017395, 3.5197830200195312, 3.536475896835327, 3.5370097160339355, 3.5748047828674316, 3.5753471851348877, 3.6131510734558105, 3.6136763095855713, 3.6286633014678955, 3.629253625869751, 3.64587664604187, 3.6464715003967285, 3.6786892414093018, 3.6792099475860596, 3.709240436553955, 3.7098002433776855, 3.725355863571167, 3.725867986679077, 3.7403647899627686, 3.7409026622772217, 3.774214506149292, 3.774817705154419, 3.8098092079162598, 3.8103256225585938, 3.8259072303771973, 3.8264193534851074, 3.840902805328369, 3.841449737548828, 3.8736634254455566, 3.8742098808288574, 3.899791717529297, 3.9003124237060547, 3.914232015609741, 3.914787530899048, 3.9297873973846436, 3.930347204208374, 3.960299491882324, 3.960885524749756, 3.9936764240264893, 3.994201421737671, 4.009188175201416, 4.0097784996032715, 4.023129463195801, 4.023663520812988, 4.051406383514404, 4.051979064941406, 4.083103179931641, 4.083632946014404, 4.095855236053467, 4.09645414352417, 4.109752655029297, 4.110342979431152, 4.137547969818115, 4.1380815505981445, 4.170334339141846, 4.170902729034424, 4.183094501495361, 4.183619499206543, 4.197513103485107, 4.198121070861816, 4.223654270172119, 4.224214553833008, 4.24753475189209, 4.248073101043701, 4.259188175201416, 4.259774208068848, 4.274804592132568, 4.275308132171631, 4.298654556274414, 4.299175262451172, 4.331432342529297, 4.3319573402404785, 4.341419219970703, 4.341948986053467, 4.35639762878418, 4.356996536254883, 4.380295276641846, 4.380855083465576, 4.406961917877197, 4.407573699951172, 4.4203081130981445, 4.420833110809326, 4.4308857917785645, 4.43141508102417, 4.454756736755371, 4.455290794372559, 4.482534885406494, 4.483055591583252, 4.4947829246521, 4.495312690734863, 4.505885601043701, 4.506402015686035, 4.528090476989746, 4.528619766235352, 4.550920486450195, 4.551427841186523, 4.561948776245117, 4.562552452087402, 4.573619842529297, 4.574210166931152, 4.59639310836792, 4.596979141235352, 4.620312690734863, 4.620846271514893, 4.630855083465576, 4.631397724151611, 4.6436285972595215, 4.644162654876709, 4.662517547607422, 4.663116455078125, 4.684735298156738, 4.685295104980469, 4.696414947509766, 4.696961879730225, 4.706944465637207, 4.707552433013916, 4.727499961853027, 4.728103160858154, 4.748645782470703, 4.7491536140441895, 4.758055686950684, 4.758650302886963, 4.770308494567871, 4.770829200744629, 4.784158229827881, 4.784765720367432, 4.793055534362793, 4.7936110496521, 4.801966190338135, 4.802560806274414, 4.813107967376709, 4.813645839691162, 4.835898399353027, 4.836414813995361, 4.860837936401367, 4.861393451690674, 4.866940498352051, 4.8675432205200195, 4.879184246063232, 4.879735469818115, 4.894739627838135, 4.895264625549316, 4.913046836853027, 4.913654327392578, 4.922486782073975, 4.923042297363281, 4.9319748878479, 4.932504653930664, 4.94749116897583, 4.948098659515381, 4.968055725097656, 4.968641757965088, 4.976393222808838, 4.976983547210693, 4.984756946563721, 4.9852776527404785, 5.00028657913208, 5.000837802886963, 5.015286445617676, 5.015829086303711, 5.02194881439209, 5.022495746612549, 5.031402111053467, 5.031931400299072, 5.046988010406494, 5.047530651092529, 5.0652995109558105, 5.065864086151123, 5.073072910308838, 5.073611259460449, 5.083094596862793, 5.083606719970703, 5.096397399902344, 5.096931457519531, 5.110269069671631, 5.110859394073486, 5.118103504180908, 5.118615627288818, 5.126397609710693, 5.1269402503967285, 5.138051509857178, 5.138589382171631, 5.154149532318115, 5.154748439788818, 5.160841941833496, 5.16137170791626, 5.168064117431641, 5.168628692626953, 5.180833339691162, 5.181397438049316, 5.189756870269775, 5.190312385559082, 5.196970462799072, 5.197482585906982, 5.2047224044799805, 5.205290794372559, 5.216944217681885, 5.217517375946045, 5.232543468475342, 5.233047008514404, 5.237534523010254, 5.238038539886475, 5.245304107666016, 5.24586820602417, 5.256397724151611, 5.25694465637207, 5.268055438995361, 5.268619537353516, 5.274739742279053, 5.275312423706055, 5.280295372009277, 5.280859470367432, 5.291380405426025, 5.29194450378418, 5.302491188049316, 5.303042411804199, 5.307473659515381, 5.308077335357666, 5.314700603485107, 5.315303802490234, 5.325811862945557, 5.326371669769287, 5.334157943725586, 5.334718227386475, 5.339188098907471, 5.3397440910339355, 5.3453168869018555, 5.345859527587891, 5.356428146362305, 5.356975078582764, 5.366375923156738, 5.366966247558594, 5.371410369873047, 5.371953010559082, 5.377521514892578, 5.378081798553467, 5.387513160705566, 5.388051509857178, 5.395837783813477, 5.396371841430664, 5.401371479034424, 5.401944637298584, 5.406414985656738, 5.406966209411621, 5.414175510406494, 5.414722442626953, 5.423055648803711, 5.423611164093018, 5.428077220916748, 5.428632736206055, 5.433042526245117, 5.433650493621826, 5.440859317779541, 5.441414833068848, 5.445850849151611, 5.446406364440918, 5.45025634765625, 5.450807094573975, 5.455824851989746, 5.456371784210205, 5.465299606323242, 5.4658331871032715, 5.474696159362793, 5.475255966186523, 5.4780731201171875, 5.478628635406494, 5.483611106872559, 5.484153747558594, 5.49028205871582, 5.490837574005127, 5.496388912200928, 5.496948719024658, 5.500807285308838, 5.5013628005981445, 5.505273818969727, 5.505829334259033, 5.51246976852417, 5.513068675994873, 5.5197529792785645, 5.520303726196289, 5.523055553436279, 5.523576259613037, 5.528611183166504, 5.529170989990234, 5.534735202789307, 5.5352864265441895, 5.540833473205566, 5.541384696960449, 5.54418420791626, 5.5447306632995605, 5.549175262451172, 5.549726486206055, 5.554166793823242, 5.554713726043701, 5.5613627433776855, 5.5619659423828125, 5.564153671264648, 5.564626693725586, 5.56635856628418, 5.568073272705078, 5.572474002838135, 5.57417106628418, 5.57525634765625, 5.576918125152588, 5.5786285400390625, 5.581913948059082, 5.58468770980835, 5.589700698852539, 5.593624114990234, 5.605278015136719, 5.611380100250244, 5.626944541931152, 5.627500057220459, 5.629726886749268, 5.6319355964660645, 5.634678840637207, 5.635286331176758, 5.658624172210693, 5.66137170791626, 5.663585186004639, 5.664144992828369, 5.666384696960449, 5.66694450378418, 5.685829162597656, 5.686380386352539, 5.688602447509766, 5.690855026245117, 5.694717884063721, 5.69526481628418, 5.719149589538574, 5.719743728637695, 5.722491264343262, 5.7230424880981445, 5.725247383117676, 5.725798606872559, 5.729166507720947, 5.729709148406982, 5.756358623504639, 5.759149551391602, 5.761362552642822, 5.761905670166016, 5.765789985656738, 5.7664103507995605, 5.798059940338135, 5.798615455627441, 5.800785541534424, 5.801397323608398, 5.803628444671631, 5.80417537689209, 5.8080339431762695, 5.808585166931152, 5.838589668273926, 5.839145183563232, 5.841388702392578, 5.841944217681885, 5.844687461853027, 5.845299243927002, 5.8497138023376465, 5.8502607345581055, 5.884730815887451, 5.885282039642334, 5.888025283813477, 5.8885722160339355, 5.890846252441406, 5.891397476196289, 5.896341323852539, 5.89694881439209, 5.931358814239502, 5.931918144226074, 5.935277938842773, 5.935829162597656, 5.938593864440918, 5.9391279220581055, 5.944118976593018, 5.9446702003479, 5.984674453735352, 5.985234260559082, 5.988003730773926, 5.9886155128479, 5.991384506225586, 5.991940498352051, 5.997465133666992, 5.9980292320251465, 6.029170989990234, 6.029717922210693, 6.033559322357178, 6.034114837646484, 6.037499904632568, 6.037999153137207, 6.043598175048828, 6.044162273406982, 6.084140777587891, 6.08468770980835, 6.08799934387207, 6.088619709014893, 6.09250020980835, 6.093047142028809, 6.0980424880981445, 6.0985894203186035, 6.136892318725586, 6.137443542480469, 6.141909599304199, 6.142456531524658, 6.145251750946045, 6.145798683166504, 6.151389122009277, 6.151936054229736, 6.195772647857666, 6.196328163146973, 6.20025634765625, 6.200794219970703, 6.204132080078125, 6.204683303833008, 6.209679126739502, 6.210226058959961, 6.24972677230835, 6.250268936157227, 6.254101753234863, 6.254709243774414, 6.25805139541626, 6.25855016708374, 6.264123439788818, 6.264665603637695, 6.307482719421387, 6.308025360107422, 6.311874866485596, 6.312443733215332, 6.316319465637207, 6.316944599151611, 6.322448253631592, 6.322999000549316, 6.362443447113037, 6.363051414489746, 6.368042469024658, 6.368597984313965, 6.372478485107422, 6.373042583465576, 6.377443790435791, 6.3780035972595215, 6.4307942390441895, 6.43135404586792, 6.436336994171143, 6.436883449554443, 6.439149379730225, 6.439704895019531, 6.445225715637207, 6.445829391479492, 6.474101543426514, 6.47471809387207, 6.479153633117676, 6.479657173156738, 6.482452392578125, 6.483007907867432, 6.489140510559082, 6.489665985107422, 6.534144878387451, 6.534687519073486, 6.53911018371582, 6.539691925048828, 6.544093132019043, 6.54469633102417, 6.549700736999512, 6.550251483917236, 6.594092845916748, 6.5946526527404785, 6.59915828704834, 6.599648475646973, 6.60300350189209, 6.603550434112549, 6.610199928283691, 6.610781192779541, 6.656905651092529, 6.657456874847412, 6.661336898803711, 6.661896705627441, 6.665229797363281, 6.665794372558594, 6.67192268371582, 6.672452449798584, 6.714653015136719, 6.7152605056762695, 6.7196831703186035, 6.720225811004639, 6.724123477935791, 6.724700450897217, 6.730790138244629, 6.731349945068359, 6.777469635009766, 6.778007984161377, 6.783563137054443, 6.784144878387451, 6.788020610809326, 6.788567543029785, 6.794110298156738, 6.794657230377197, 6.838003635406494, 6.838550090789795, 6.844097137451172, 6.844666004180908, 6.848567962646484, 6.849123477935791, 6.854687690734863, 6.855199813842773, 6.906319618225098, 6.906931400299072, 6.910794258117676, 6.9113240242004395, 6.9152607917785645, 6.915824890136719, 6.9213409423828125, 6.921901226043701, 6.963029384613037, 6.9635329246521, 6.969644069671631, 6.9702606201171875, 6.97358512878418, 6.974123477935791, 6.980238914489746, 6.980772495269775, 7.028563499450684, 7.029114723205566, 7.034674644470215, 7.035212993621826, 7.039118766784668, 7.039679050445557, 7.045742034912109, 7.046354293823242, 7.092981815338135, 7.093545913696289, 7.097469806671143, 7.098033905029297, 7.101888179779053, 7.102456569671631, 7.1085591316223145, 7.109097003936768, 7.15802526473999, 7.158567905426025, 7.1624650955200195, 7.163007736206055, 7.166853427886963, 7.167452335357666, 7.173524379730225, 7.174079895019531, 7.218520164489746, 7.219079971313477, 7.223515510559082, 7.224127769470215, 7.227968692779541, 7.228515625, 7.234639644622803, 7.235195159912109, 7.284127712249756, 7.284626483917236, 7.2890625, 7.289674282073975, 7.293554782867432, 7.2941060066223145, 7.299639701843262, 7.3001909255981445, 7.3446831703186035, 7.34523868560791, 7.349110126495361, 7.349661350250244, 7.3535284996032715, 7.35408878326416, 7.360225677490234, 7.360759735107422, 7.412981986999512, 7.4135284423828125, 7.417417526245117, 7.41796875, 7.421288967132568, 7.421909809112549, 7.427951335906982, 7.428524494171143, 7.466875076293945, 7.4674391746521, 7.471861839294434, 7.47243070602417, 7.475733757019043, 7.476315498352051, 7.481280326843262, 7.481883525848389, 7.5274481773376465, 7.527990341186523, 7.53131103515625, 7.531861782073975, 7.535725116729736, 7.536341190338135, 7.541879653930664, 7.542426586151123, 7.584613800048828, 7.585195064544678, 7.58966588973999, 7.590204238891602, 7.593511581420898, 7.594058513641357, 7.6007466316223145, 7.601293563842773, 7.645724773406982, 7.6462931632995605, 7.650191307067871, 7.6507248878479, 7.654644012451172, 7.655195236206055, 7.660772323608398, 7.6613240242004395, 7.700763702392578, 7.701310634613037, 7.705173492431641, 7.705755233764648, 7.709661483764648, 7.710216999053955, 7.715169429779053, 7.715729236602783, 7.757382869720459, 7.757938385009766, 7.76242208480835, 7.762986183166504, 7.766275882720947, 7.766892433166504, 7.772395610809326, 7.772947311401367, 7.812426567077637, 7.812982082366943, 7.816293239593506, 7.816840171813965, 7.82077693939209, 7.82135009765625, 7.825733661651611, 7.826284885406494, 7.8690667152404785, 7.869609355926514, 7.873554706573486, 7.874062538146973, 7.877408981323242, 7.877982139587402, 7.882981777191162, 7.883493900299072, 7.915182113647461, 7.915750980377197, 7.919066905975342, 7.919657230377197, 7.923563480377197, 7.92406702041626, 7.927951335906982, 7.928502559661865, 7.967973232269287, 7.968520164489746, 7.971831798553467, 7.972382545471191, 7.975733757019043, 7.976271629333496, 7.980751037597656, 7.981271743774414, 8.0157470703125, 8.016310691833496, 8.02021312713623, 8.020768165588379, 8.023507118225098, 8.02404499053955, 8.029106140136719, 8.02966594696045, 8.069084167480469, 8.069640159606934, 8.07294750213623, 8.073507308959961, 8.075733184814453, 8.076272010803223, 8.080755233764648, 8.081340789794922, 8.112981796264648, 8.113533020019531, 8.117383003234863, 8.117942810058594, 8.121306419372559, 8.121835708618164, 8.126297950744629, 8.12685775756836, 8.162430763244629, 8.162999153137207, 8.166319847106934, 8.16683578491211, 8.169036865234375, 8.169622421264648, 8.174626350402832, 8.175182342529297, 8.206315040588379, 8.206857681274414, 8.210746765136719, 8.211258888244629, 8.21349811553955, 8.214049339294434, 8.218493461608887, 8.219049453735352, 8.257943153381348, 8.258528709411621, 8.26128101348877, 8.261848449707031, 8.264066696166992, 8.264604568481445, 8.268511772155762, 8.269071578979492, 8.290763854980469, 8.291275978088379, 8.295204162597656, 8.295720100402832, 8.297964096069336, 8.2985200881958, 8.302421569824219, 8.302982330322266, 8.335186958312988, 8.335750579833984, 8.337960243225098, 8.338515281677246, 8.341259002685547, 8.341818809509277, 8.345186233520508, 8.345725059509277, 8.376315116882324, 8.376827239990234, 8.379600524902344, 8.380151748657227, 8.382977485656738, 8.383515357971191, 8.38744831085205, 8.387955665588379, 8.41852855682373, 8.419088363647461, 8.422426223754883, 8.424617767333984, 8.428498268127441, 8.429062843322754, 8.4557466506958, 8.456258773803711, 8.459071159362793, 8.459613800048828, 8.461840629577637, 8.46239185333252, 8.465737342834473, 8.46629810333252, 8.495173454284668, 8.495759963989258, 8.49854564666748, 8.499093055725098, 8.501301765441895, 8.501861572265625, 8.50570297241211, 8.506293296813965, 8.531314849853516, 8.531826972961426, 8.53464412689209, 8.535208702087402, 8.536866188049316, 8.537425994873047, 8.540738105773926, 8.541319847106934, 8.571836471557617, 8.572391510009766, 8.574630737304688, 8.577425956726074, 8.580741882324219, 8.581275939941406, 8.602981567382812, 8.603541374206543, 8.605742454528809, 8.608537673950195, 8.611809730529785, 8.612443923950195, 8.63681411743164, 8.639631271362305, 8.642378807067871, 8.642942428588867, 8.645707130432129, 8.646254539489746, 8.66848087310791, 8.671293258666992, 14.144461631774902]),
                       array.array('f', [13.090999603271484, 14.199999809265137, 14.199999809265137, 13.668999671936035, 13.581999778747559, 13.52400016784668, 13.52400016784668, 13.49899959564209, 13.49899959564209, 13.491999626159668, 13.491999626159668, 13.470000267028809, 13.470000267028809, 13.463000297546387, 13.463000297546387, 13.437999725341797, 13.437999725341797, 13.409000396728516, 13.409000396728516, 13.380000114440918, 13.380000114440918, 13.354000091552734, 13.354000091552734, 13.347000122070312, 13.347000122070312, 13.326000213623047, 12.776000022888184, 12.368000030517578, 12.255999565124512, 12.255999565124512, 12.248000144958496, 12.248000144958496, 12.234000205993652, 12.234000205993652, 12.241000175476074, 12.241000175476074, 12.248000144958496, 12.248000144958496, 12.241000175476074, 12.241000175476074, 12.234000205993652, 12.234000205993652, 12.22700023651123, 12.22700023651123, 12.218999862670898, 12.218999862670898, 12.211999893188477, 12.211999893188477, 12.204999923706055, 12.204999923706055, 12.197999954223633, 12.197999954223633, 12.190999984741211, 12.190999984741211, 12.182999610900879, 12.182999610900879, 12.175999641418457, 12.175999641418457, 12.168999671936035, 12.168999671936035, 12.161999702453613, 12.161999702453613, 12.154000282287598, 12.154000282287598, 12.147000312805176, 12.147000312805176, 12.140000343322754, 12.140000343322754, 12.133000373840332, 12.133000373840332, 12.125, 12.125, 12.118000030517578, 12.118000030517578, 12.111000061035156, 12.111000061035156, 12.104000091552734, 12.104000091552734, 12.097000122070312, 12.097000122070312, 12.08899974822998, 12.08899974822998, 12.081999778747559, 12.081999778747559, 12.074999809265137, 12.074999809265137, 12.067999839782715, 12.067999839782715, 12.0600004196167, 12.0600004196167, 12.053000450134277, 12.053000450134277, 12.045999526977539, 12.045999526977539, 12.038999557495117, 12.038999557495117, 12.031000137329102, 12.031000137329102, 12.02400016784668, 12.02400016784668, 12.017000198364258, 12.017000198364258, 12.010000228881836, 12.010000228881836, 12.003000259399414, 12.003000259399414, 11.994999885559082, 11.994999885559082, 11.98799991607666, 11.98799991607666, 11.980999946594238, 11.980999946594238, 11.973999977111816, 11.973999977111816, 11.965999603271484, 11.965999603271484, 11.958999633789062, 11.958999633789062, 11.95199966430664, 11.95199966430664, 11.944999694824219, 11.944999694824219, 11.937000274658203, 11.937000274658203, 11.930000305175781, 11.930000305175781, 11.92300033569336, 11.92300033569336, 11.916000366210938, 11.916000366210938, 11.909000396728516, 11.909000396728516, 11.901000022888184, 11.901000022888184, 11.894000053405762, 11.894000053405762, 11.88700008392334, 11.88700008392334, 11.880000114440918, 11.880000114440918, 11.871999740600586, 11.871999740600586, 11.864999771118164, 11.864999771118164, 11.857999801635742, 11.857999801635742, 11.85099983215332, 11.85099983215332, 11.843000411987305, 11.843000411987305, 11.836000442504883, 11.836000442504883, 11.829000473022461, 11.829000473022461, 11.821999549865723, 11.821999549865723, 11.8149995803833, 11.8149995803833, 11.807000160217285, 11.807000160217285, 11.800000190734863, 11.800000190734863, 11.793000221252441, 11.793000221252441, 11.78600025177002, 11.78600025177002, 11.777999877929688, 11.777999877929688, 11.770999908447266, 11.770999908447266, 11.763999938964844, 11.763999938964844, 11.756999969482422, 11.756999969482422, 11.75, 11.75, 11.741999626159668, 11.741999626159668, 11.734999656677246, 11.734999656677246, 11.727999687194824, 11.727999687194824, 11.720999717712402, 11.720999717712402, 11.713000297546387, 11.713000297546387, 11.706000328063965, 11.706000328063965, 11.699000358581543, 11.699000358581543, 11.692000389099121, 11.692000389099121, 11.684000015258789, 11.684000015258789, 11.677000045776367, 11.677000045776367, 11.670000076293945, 11.670000076293945, 11.663000106811523, 11.663000106811523, 11.656000137329102, 11.656000137329102, 11.64799976348877, 11.64799976348877, 11.640999794006348, 11.640999794006348, 11.633999824523926, 11.633999824523926, 11.626999855041504, 11.626999855041504, 11.619000434875488, 11.619000434875488, 11.612000465393066, 11.612000465393066, 11.604999542236328, 11.604999542236328, 11.597999572753906, 11.597999572753906, 11.59000015258789, 11.59000015258789, 11.583000183105469, 11.583000183105469, 11.576000213623047, 11.576000213623047, 11.569000244140625, 11.569000244140625, 11.562000274658203, 11.562000274658203, 11.553999900817871, 11.553999900817871, 11.54699993133545, 11.54699993133545, 11.539999961853027, 11.539999961853027, 11.532999992370605, 11.532999992370605, 11.524999618530273, 11.524999618530273, 11.517999649047852, 11.517999649047852, 11.51099967956543, 11.51099967956543, 11.503999710083008, 11.503999710083008, 11.496000289916992, 11.496000289916992, 11.48900032043457, 11.48900032043457, 11.482000350952148, 11.482000350952148, 11.475000381469727, 11.475000381469727, 11.468000411987305, 11.468000411987305, 11.460000038146973, 11.460000038146973, 11.45300006866455, 11.45300006866455, 11.446000099182129, 11.446000099182129, 11.439000129699707, 11.439000129699707, 11.430999755859375, 11.430999755859375, 11.423999786376953, 11.423999786376953, 11.416999816894531, 11.416999816894531, 11.40999984741211, 11.40999984741211, 11.402000427246094, 11.402000427246094, 11.395000457763672, 11.395000457763672, 11.387999534606934, 11.387999534606934, 11.380999565124512, 11.380999565124512, 11.37399959564209, 11.37399959564209, 11.366000175476074, 11.366000175476074, 11.359000205993652, 11.359000205993652, 11.35200023651123, 11.35200023651123, 11.345000267028809, 11.345000267028809, 11.336999893188477, 11.336999893188477, 11.329999923706055, 11.329999923706055, 11.322999954223633, 11.322999954223633, 11.315999984741211, 11.315999984741211, 11.309000015258789, 11.309000015258789, 11.300999641418457, 11.300999641418457, 11.293999671936035, 11.293999671936035, 11.286999702453613, 11.286999702453613, 11.279999732971191, 11.279999732971191, 11.272000312805176, 11.272000312805176, 11.265000343322754, 11.265000343322754, 11.258000373840332, 11.258000373840332, 11.25100040435791, 11.25100040435791, 11.243000030517578, 11.243000030517578, 11.236000061035156, 11.236000061035156, 11.229000091552734, 11.229000091552734, 11.222000122070312, 11.222000122070312, 11.21500015258789, 11.21500015258789, 11.206999778747559, 11.206999778747559, 11.199999809265137, 11.199999809265137, 11.192999839782715, 11.192999839782715, 11.185999870300293, 11.185999870300293, 11.178000450134277, 11.178000450134277, 11.170999526977539, 11.170999526977539, 11.163999557495117, 11.163999557495117, 11.156999588012695, 11.156999588012695, 11.14900016784668, 11.14900016784668, 11.142000198364258, 11.142000198364258, 11.135000228881836, 11.135000228881836, 11.128000259399414, 11.128000259399414, 11.121000289916992, 11.121000289916992, 11.11299991607666, 11.11299991607666, 11.105999946594238, 11.105999946594238, 11.098999977111816, 11.098999977111816, 11.092000007629395, 11.092000007629395, 11.083999633789062, 11.083999633789062, 11.07699966430664, 11.07699966430664, 11.069999694824219, 11.069999694824219, 11.062999725341797, 11.062999725341797, 11.055000305175781, 11.055000305175781, 11.04800033569336, 11.04800033569336, 11.041000366210938, 11.041000366210938, 11.034000396728516, 11.034000396728516, 11.027000427246094, 11.027000427246094, 11.019000053405762, 11.019000053405762, 11.01200008392334, 11.01200008392334, 11.005000114440918, 11.005000114440918, 10.998000144958496, 10.998000144958496, 10.989999771118164, 10.989999771118164, 10.982999801635742, 10.982999801635742, 10.97599983215332, 10.97599983215332, 10.968999862670898, 10.968999862670898, 10.961000442504883, 10.961000442504883, 10.954000473022461, 10.954000473022461, 10.946999549865723, 10.946999549865723, 10.9399995803833, 10.9399995803833, 10.932999610900879, 10.932999610900879, 10.925000190734863, 10.925000190734863, 10.918000221252441, 10.918000221252441, 10.91100025177002, 10.91100025177002, 10.904000282287598, 10.904000282287598, 10.895999908447266, 10.895999908447266, 10.888999938964844, 10.888999938964844, 10.881999969482422, 10.881999969482422, 10.875, 10.875, 10.868000030517578, 10.868000030517578, 10.859999656677246, 10.859999656677246, 10.852999687194824, 10.852999687194824, 10.845999717712402, 10.845999717712402, 10.83899974822998, 10.83899974822998, 10.831000328063965, 10.831000328063965, 10.824000358581543, 10.824000358581543, 10.817000389099121, 10.817000389099121, 10.8100004196167, 10.8100004196167, 10.802000045776367, 10.802000045776367, 10.795000076293945, 10.795000076293945, 10.788000106811523, 10.788000106811523, 10.781000137329102, 10.781000137329102, 10.77400016784668, 10.77400016784668, 10.765999794006348, 10.765999794006348, 10.758999824523926, 10.758999824523926, 10.751999855041504, 10.751999855041504, 10.744999885559082, 10.744999885559082, 10.737000465393066, 10.737000465393066, 10.729999542236328, 10.729999542236328, 10.722999572753906, 10.722999572753906, 10.715999603271484, 10.715999603271484, 10.708000183105469, 10.708000183105469, 10.701000213623047, 10.701000213623047, 10.694000244140625, 10.694000244140625, 10.687000274658203, 10.687000274658203, 10.680000305175781, 10.680000305175781, 10.67199993133545, 10.67199993133545, 10.664999961853027, 10.664999961853027, 10.657999992370605, 10.657999992370605, 10.651000022888184, 10.651000022888184, 10.642999649047852, 10.642999649047852, 10.63599967956543, 10.63599967956543, 10.628999710083008, 10.628999710083008, 10.621999740600586, 10.621999740600586, 10.61400032043457, 10.61400032043457, 10.607000350952148, 10.607000350952148, 10.600000381469727, 10.600000381469727, 10.593000411987305, 10.593000411987305, 10.586000442504883, 10.586000442504883, 10.57800006866455, 10.57800006866455, 10.571000099182129, 10.571000099182129, 10.564000129699707, 10.564000129699707, 10.557000160217285, 10.557000160217285, 10.548999786376953, 10.548999786376953, 10.541999816894531, 10.541999816894531, 10.53499984741211, 10.53499984741211, 10.527999877929688, 10.527999877929688, 10.520000457763672, 10.520000457763672, 10.512999534606934, 10.512999534606934, 10.505999565124512, 10.505999565124512, 10.49899959564209, 12.07800006866455, 12.190999984741211, 12.303000450134277, 12.3100004196167, 12.335000038146973, 12.335000038146973, 12.36400032043457, 12.36400032043457, 12.392999649047852, 12.392999649047852, 12.42199993133545, 12.42199993133545, 12.451000213623047, 12.451000213623047, 12.458000183105469, 12.458000183105469, 12.472000122070312, 12.472000122070312, 12.479999542236328, 12.479999542236328, 12.494000434875488, 12.494000434875488, 12.50100040435791, 12.50100040435791, 12.508999824523926, 12.508999824523926, 12.515999794006348, 12.515999794006348, 12.529999732971191, 12.529999732971191, 12.538000106811523, 12.538000106811523, 12.545000076293945, 12.545000076293945, 12.552000045776367, 12.552000045776367, 12.559000015258789, 12.559000015258789, 12.565999984741211, 12.565999984741211, 12.581000328063965, 12.581000328063965, 12.588000297546387, 12.588000297546387, 12.595000267028809, 12.595000267028809, 12.602999687194824, 12.602999687194824, 12.609999656677246, 12.609999656677246, 12.616999626159668, 12.616999626159668, 12.62399959564209, 12.62399959564209, 12.631999969482422, 12.631999969482422, 12.638999938964844, 12.638999938964844, 12.645999908447266, 12.645999908447266, 12.652999877929688, 12.652999877929688, 12.65999984741211, 12.65999984741211, 12.668000221252441, 12.668000221252441, 12.675000190734863, 12.675000190734863, 12.682000160217285, 12.682000160217285, 12.689000129699707, 12.689000129699707, 12.696999549865723, 12.696999549865723, 12.704000473022461, 12.704000473022461, 12.711000442504883, 12.711000442504883, 12.718000411987305, 12.718000411987305, 12.725000381469727, 12.725000381469727, 12.732999801635742, 12.732999801635742, 12.739999771118164, 12.739999771118164, 12.746999740600586, 12.746999740600586, 12.753999710083008, 12.753999710083008, 12.76200008392334, 12.76200008392334, 12.769000053405762, 12.769000053405762, 12.776000022888184, 12.776000022888184, 12.782999992370605, 12.782999992370605, 12.791000366210938, 12.791000366210938, 12.79800033569336, 12.79800033569336, 12.805000305175781, 12.805000305175781, 12.812000274658203, 12.812000274658203, 12.819000244140625, 12.819000244140625, 12.82699966430664, 12.82699966430664, 12.833999633789062, 12.833999633789062, 12.840999603271484, 12.840999603271484, 12.847999572753906, 12.847999572753906, 12.855999946594238, 12.855999946594238, 12.86299991607666, 12.86299991607666, 12.869999885559082, 12.869999885559082, 12.876999855041504, 12.876999855041504, 12.885000228881836, 12.885000228881836, 12.892000198364258, 12.892000198364258, 12.89900016784668, 12.89900016784668, 12.906000137329102, 12.906000137329102, 12.913000106811523, 12.913000106811523, 12.920999526977539, 12.920999526977539, 12.928000450134277, 12.928000450134277, 12.9350004196167, 12.9350004196167, 12.942000389099121, 12.942000389099121, 12.949999809265137, 12.949999809265137, 12.956999778747559, 12.956999778747559, 12.96399974822998, 12.96399974822998, 12.970999717712402, 12.970999717712402, 12.979000091552734, 12.979000091552734, 12.986000061035156, 12.986000061035156, 12.993000030517578, 12.993000030517578, 13.0, 13.0, 13.006999969482422, 13.006999969482422, 13.015000343322754, 13.015000343322754, 13.022000312805176, 13.022000312805176, 13.029000282287598, 13.029000282287598, 13.03600025177002, 13.03600025177002, 13.043999671936035, 13.043999671936035, 13.050999641418457, 13.050999641418457, 13.057999610900879, 13.057999610900879, 13.0649995803833, 13.0649995803833, 13.072999954223633, 13.072999954223633, 13.079999923706055, 13.079999923706055, 13.086999893188477, 13.086999893188477, 13.093999862670898, 13.093999862670898, 13.10099983215332, 13.10099983215332, 13.109000205993652, 13.109000205993652, 13.116000175476074, 13.116000175476074, 13.123000144958496, 13.123000144958496, 13.130000114440918, 13.130000114440918, 13.137999534606934, 13.137999534606934, 13.145000457763672, 13.145000457763672, 13.152000427246094, 13.152000427246094, 13.159000396728516, 13.159000396728516, 13.166000366210938, 13.166000366210938, 13.173999786376953, 13.173999786376953, 13.180999755859375, 13.180999755859375, 13.187999725341797, 13.187999725341797, 13.194999694824219, 13.194999694824219, 13.20300006866455, 13.20300006866455, 13.210000038146973, 13.210000038146973, 13.217000007629395, 13.217000007629395, 13.223999977111816, 13.223999977111816, 13.232000350952148, 13.232000350952148, 13.23900032043457, 13.23900032043457, 13.246000289916992, 13.246000289916992, 13.253000259399414, 13.253000259399414, 13.260000228881836, 13.260000228881836, 13.267999649047852, 13.267999649047852, 13.274999618530273, 13.274999618530273, 13.281999588012695, 13.281999588012695, 13.288999557495117, 13.288999557495117, 13.29699993133545, 13.29699993133545, 13.303999900817871, 13.303999900817871, 13.310999870300293, 13.310999870300293, 13.317999839782715, 13.317999839782715, 13.326000213623047, 13.326000213623047, 13.333000183105469, 13.333000183105469, 13.34000015258789, 13.34000015258789, 13.347000122070312, 13.347000122070312, 13.354000091552734, 13.354000091552734, 13.362000465393066, 13.362000465393066, 13.369000434875488, 13.369000434875488, 13.37600040435791, 13.37600040435791, 13.383000373840332, 13.383000373840332, 13.390999794006348, 13.390999794006348, 13.39799976348877, 13.39799976348877, 13.404999732971191, 13.404999732971191, 13.411999702453613, 13.411999702453613, 13.420000076293945, 13.420000076293945, 13.427000045776367, 13.427000045776367, 13.434000015258789, 13.434000015258789, 13.440999984741211, 13.440999984741211, 13.447999954223633, 13.447999954223633, 13.456000328063965, 13.456000328063965, 13.463000297546387, 13.463000297546387, 13.470000267028809, 13.470000267028809, 13.47700023651123, 13.47700023651123, 13.484999656677246, 13.484999656677246, 13.491999626159668, 13.491999626159668, 13.49899959564209, 13.49899959564209, 13.505999565124512, 13.505999565124512, 13.513999938964844, 13.513999938964844, 13.520999908447266, 13.520999908447266, 13.527999877929688, 13.527999877929688, 13.53499984741211, 13.53499984741211, 13.541999816894531, 13.541999816894531, 13.550000190734863, 13.550000190734863, 13.557000160217285, 13.557000160217285, 13.564000129699707, 13.564000129699707, 13.571000099182129, 13.571000099182129, 13.579000473022461, 13.579000473022461, 13.586000442504883, 13.586000442504883, 13.593000411987305, 13.593000411987305, 13.600000381469727, 13.600000381469727, 13.607999801635742, 13.607999801635742, 13.614999771118164, 13.614999771118164, 13.621999740600586, 13.621999740600586, 13.628999710083008, 13.628999710083008, 13.63599967956543, 13.63599967956543, 13.644000053405762, 13.644000053405762, 13.651000022888184, 13.651000022888184, 13.657999992370605, 13.657999992370605, 13.664999961853027, 13.664999961853027, 13.67300033569336, 13.67300033569336, 13.680000305175781, 13.680000305175781, 13.687000274658203, 13.687000274658203, 13.694000244140625, 13.694000244140625, 13.701000213623047, 13.701000213623047, 13.708999633789062, 13.708999633789062, 13.715999603271484, 13.715999603271484, 13.722999572753906, 13.722999572753906, 13.729999542236328, 13.729999542236328, 13.73799991607666, 13.73799991607666, 13.744999885559082, 13.744999885559082, 13.751999855041504, 13.751999855041504, 13.758999824523926, 13.758999824523926, 13.767000198364258, 13.767000198364258, 13.77400016784668, 13.77400016784668, 13.781000137329102, 13.781000137329102, 13.788000106811523, 13.788000106811523, 13.795000076293945, 13.795000076293945, 13.803000450134277, 13.803000450134277, 13.8100004196167, 13.8100004196167, 13.817000389099121, 13.817000389099121, 13.824000358581543, 13.824000358581543, 13.831999778747559, 13.831999778747559, 13.83899974822998, 13.83899974822998, 13.845999717712402, 13.845999717712402, 13.852999687194824, 13.852999687194824, 13.861000061035156, 13.861000061035156, 13.868000030517578, 13.868000030517578, 13.875, 13.875, 13.881999969482422, 13.881999969482422, 13.888999938964844, 13.888999938964844, 13.897000312805176, 13.897000312805176, 13.904000282287598, 13.904000282287598, 13.91100025177002, 13.91100025177002, 13.918000221252441, 13.918000221252441, 13.925999641418457, 13.925999641418457, 13.932999610900879, 13.932999610900879, 13.9399995803833, 13.9399995803833, 13.946999549865723, 13.946999549865723, 13.954999923706055, 13.954999923706055, 13.961999893188477, 13.961999893188477, 13.968999862670898, 13.968999862670898, 13.97599983215332, 13.97599983215332, 13.982999801635742, 13.982999801635742, 13.991000175476074, 13.991000175476074, 14.005000114440918, 14.005000114440918, 14.01200008392334, 14.01200008392334, 14.020000457763672, 14.020000457763672, 14.027000427246094, 14.027000427246094, 14.034000396728516, 14.034000396728516, 14.041000366210938, 14.041000366210938, 14.048999786376953, 14.048999786376953, 14.055999755859375, 14.055999755859375, 14.062999725341797, 14.062999725341797, 14.069999694824219, 14.069999694824219, 14.07699966430664, 14.07699966430664, 14.085000038146973, 14.085000038146973, 14.092000007629395, 14.092000007629395, 14.098999977111816, 14.098999977111816, 14.105999946594238, 14.105999946594238, 14.121000289916992, 14.121000289916992, 14.128000259399414, 14.128000259399414, 14.135000228881836, 14.135000228881836, 14.149999618530273, 14.149999618530273, 14.156999588012695, 14.156999588012695, 14.170999526977539, 14.170999526977539, 14.178999900817871, 14.178999900817871, 14.185999870300293, 14.185999870300293, 14.199999809265137, 14.199999809265137])
                       ),
                      (array.array('f', [0.00021722221572417766, 0.0004602777771651745, 0.0010158333461731672, 0.0016233333153650165, 0.0021744444966316223, 0.002730000065639615, 0.003285555634647608, 0.003836944466456771, 0.004388055764138699, 0.004939444363117218, 0.005495000164955854, 0.0060461112298071384, 0.006588611286133528, 0.007135555613785982, 0.007691110949963331, 0.008246666751801968, 0.00885861087590456, 0.00940111093223095, 0.009956666268408298, 0.010507777333259583, 0.011067777872085571, 0.01162333320826292, 0.012174444273114204, 0.012729999609291553, 0.013285555876791477, 0.013832500204443932, 0.01438805554062128, 0.01493055559694767, 0.015490555204451084, 0.016046110540628433, 0.016588611528277397, 0.01714416593313217, 0.017699722200632095, 0.018263889476656914, 0.01882833242416382, 0.019375000149011612, 0.01992194354534149, 0.020477499812841415, 0.02103305608034134, 0.02158416621387005, 0.022135555744171143, 0.022691112011671066, 0.023298611864447594, 0.02379361167550087, 0.024353332817554474, 0.024908889085054398, 0.02552083320915699, 0.026019999757409096, 0.02662777714431286, 0.027178889140486717, 0.027738889679312706, 0.028289999812841415, 0.02884555608034134, 0.029401110485196114, 0.029956666752696037, 0.03051222302019596, 0.03106333315372467, 0.031618889421224594, 0.03217027708888054, 0.03272583335638046, 0.03327277675271034, 0.033828333020210266, 0.03438388928771019, 0.034939445555210114, 0.03549500182271004, 0.0360461100935936, 0.03661472350358963, 0.0371788889169693, 0.03773888945579529, 0.0382813885807991, 0.038836944848299026, 0.03940555453300476, 0.039904721081256866, 0.040486112236976624, 0.04105055704712868, 0.04161027818918228, 0.04215722158551216, 0.042725834995508194, 0.0432855561375618, 0.04383249953389168, 0.04439250007271767, 0.0449611097574234, 0.045507777482271194, 0.04605916514992714, 0.046610277146101, 0.04716583341360092, 0.047717221081256866, 0.04827277734875679, 0.04882388934493065, 0.04937500134110451, 0.04992194473743439, 0.050473056733608246, 0.051033057272434235, 0.051584165543317795, 0.05213972181081772, 0.05270416662096977, 0.05326833203434944, 0.05383250117301941, 0.0543925017118454, 0.05494360998272896]),
                       array.array('f', [13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.10099983215332, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.10099983215332, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.097000122070312, 13.097000122070312, 13.10099983215332, 13.10099983215332, 13.10099983215332, 13.097000122070312, 13.097000122070312])
                       ),
                      (array.array('f', [0.00021722221572417766, 0.0005252778064459562, 0.0010763888712972403, 0.0016233333153650165, 0.0021744444966316223, 0.002730000065639615, 0.003285555634647608, 0.003832499962300062, 0.004388055764138699, 0.004939444363117218, 0.0055555556900799274, 0.006102500017732382, 0.006649444345384836, 0.007196388673037291, 0.007747500203549862, 0.008298611268401146, 0.00885861087590456, 0.00940111093223095, 0.009956666268408298, 0.010507777333259583, 0.011059166863560677, 0.011614722199738026, 0.012222222052514553, 0.012721388600766659, 0.013285555876791477, 0.013832500204443932, 0.01438805554062128, 0.014995832927525043, 0.015546943992376328, 0.016093889251351357, 0.01664944365620613, 0.017204999923706055, 0.017764722928404808, 0.018329165875911713, 0.018888888880610466, 0.019435834139585495, 0.019986944273114204, 0.0205383338034153, 0.021093888208270073, 0.02165805548429489, 0.022213611751794815, 0.022717222571372986, 0.02332916669547558, 0.02388888970017433, 0.024383889511227608, 0.0249958336353302, 0.02554694376885891, 0.026098055765032768, 0.026653612032532692, 0.027204999700188637, 0.027751944959163666, 0.028303055092692375, 0.028867222368717194, 0.029418611899018288, 0.029982777312397957, 0.030533889308571815, 0.03108944371342659, 0.03164083510637283, 0.03219194337725639, 0.03275194391608238, 0.033311665058135986, 0.03386722132563591, 0.034422777593135834, 0.034969720989465714, 0.0355297215282917, 0.03608083352446556, 0.036636389791965485, 0.03718750178813934, 0.03774305433034897, 0.03829444572329521, 0.03884555399417877, 0.03941861167550087, 0.03993944451212883, 0.040499165654182434, 0.04105916619300842, 0.0416061095893383, 0.042222220450639725, 0.042725834995508194, 0.04328138753771782, 0.04383249953389168, 0.0443880558013916, 0.04496527835726738, 0.045516666024923325, 0.04607222229242325, 0.04661889001727104, 0.047174446284770966, 0.04772583395242691, 0.048285555094480515, 0.04884111136198044, 0.04939666762948036, 0.04994805529713631, 0.05050777643918991, 0.05105472356081009, 0.05166250094771385, 0.05221361294388771, 0.05276472121477127, 0.05332472175359726, 0.05383250117301941, 0.0543925017118454, 0.05494360998272896]),
                       array.array('f', [13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.093000411987305, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.093000411987305, 13.097000122070312, 13.093000411987305, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.093000411987305, 13.093000411987305, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.093000411987305, 13.097000122070312, 13.093000411987305, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.093000411987305, 13.097000122070312, 13.097000122070312, 13.093000411987305, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.093000411987305, 13.097000122070312, 13.097000122070312, 13.093000411987305, 13.097000122070312, 13.093000411987305, 13.097000122070312, 13.093000411987305, 13.093000411987305, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.097000122070312, 13.093000411987305, 13.097000122070312, 13.097000122070312, 13.093000411987305, 13.093000411987305, 13.097000122070312])
                       ),
                      ]

            fig = plt.figure(figsize=(15, 8))
            epsilon = 0.03
            threshold = 0.02
            for sub, (x, y) in zip((221, 222, 223, 224), series):
                resampled_series = RdpDecimateSeries(chunk_size=20, epsilon=epsilon, threshold=threshold)
                resampled_series.extend(x, y)
                resampled_series.force_processing()
                ax = fig.add_subplot(sub)
                ax.plot(x, y, 'r', label=f'{len(x)}')
                ax.plot(resampled_series.x, resampled_series.y, 'k', label=f'{len(resampled_series)}')
                ax.set_title(f'{len(x)} - {len(x) - len(resampled_series)} = {len(resampled_series)} '
                             f'(eps = {epsilon}, tsh = {threshold})')

            plt.subplots_adjust(left=0.1, bottom=0.1, right=0.9, top=0.9, wspace=0.4, hspace=0.4)
            plt.show()

        def create_sin_series(resampled_series):
            for i, j in zip(range(100), list(np.linspace(0, 2 * np.pi, 100))):
                resampled_series.append(i, math.sin(j))

        def create_linear_series(resampled_series):
            for i in range(100):
                resampled_series.append(i, i + 1000)

        def decimate_series():
            test = ((create_linear_series, array.array('f', [0.0, 99.0])),
                    (create_sin_series, array.array('f', [0.0, 6.0, 11.0, 15.0, 19.0, 21.0, 24.0, 27.0, 30.0, 33.0,
                                                          36.0, 39.0, 43.0, 54.0, 59.0, 61.0, 64.0, 67.0, 70.0, 73.0,
                                                          76.0, 79.0, 81.0, 83.0, 86.0, 89.0, 92.0, 99.0])),)
            for create_series, result in test:
                resampled_series = RdpDecimateSeries(chunk_size=20, epsilon=0.01)
                create_series(resampled_series)
                assert resampled_series.x == result

        async def decimate_series_task():
            from concurrent.futures import ProcessPoolExecutor

            test = ((create_linear_series, array.array('f', [0.0, 99.0])),
                    (create_sin_series, array.array('f', [0.0, 6.0, 11.0, 15.0, 20.0, 24.0, 28.0, 30.0, 33.0,
                                                          37.0, 41.0, 45.0, 55.0, 61.0, 64.0, 67.0, 70.0, 72.0,
                                                          75.0, 78.0, 81.0, 84.0, 87.0, 90.0, 94.0, 99.0])),)
            pool = ProcessPoolExecutor(max_workers=1)
            for create_series, result in test:
                resampled_series = RdpDecimateSeries(chunk_size=20, epsilon=0.01,
                                                     loop=asyncio.get_event_loop(), pool=pool)
                create_series(resampled_series)
                await resampled_series.processing_done()
                assert resampled_series.x == result

        factors_test()
        asyncio.run(decimate_series_task())
        decimate_series()

    _tests()
